"""
implementation of multi factor ewm model
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Dict, Optional
from enum import Enum

# qis
import qis.utils.df_ops as dfo
import qis.plots.time_series as pts
import qis.models.linear.ewm as ewm
from qis.models.linear.ewm import MeanAdjType, InitType
from qis import TimePeriod


class LinearModel:
    """
    core class to store and process of data for different linear signals of shape:
    y =  loadings^T @ x
    """
    def __init__(self,
                 x: pd.DataFrame,  # t, x_n factors
                 y: pd.DataFrame,  # t, y_m factors
                 loadings: Dict[str, pd.DataFrame],
                 ):

        str_factors = list(loadings.keys())
        assert str_factors == x.columns.to_list()

        self.x = x
        self.y = y
        self.loadings = loadings

    def print(self):
        print(f"x:\n{self.x}")
        print(f"y:\n{self.y}")
        for factor, loading in self.loadings.items():
            print(f"{factor}:\n{loading}")

    def compute_agg_factor_exposures(self,
                                     asset_exposures: pd.DataFrame
                                     ) -> pd.DataFrame:
        factor_exposures = {}
        for factor, loading in self.loadings.items():
            factor_exposures[factor] = loading.multiply(asset_exposures).sum(axis=1)
        factor_exposures = pd.DataFrame.from_dict(factor_exposures)
        return factor_exposures

    def plot_factor_loadings(self,
                             factor: str,
                             var_format: str = '{:,.2f}',
                             time_period: TimePeriod = None,
                             ax: plt.Subplot = None,
                             **kwargs):
        df = self.loadings[factor]
        if time_period is not None:
            df = time_period.locate(df)
        pts.plot_time_series(df=df,
                             var_format=var_format,
                             ax=ax, **kwargs)


def estimate_ewm_linear_model(x: pd.DataFrame,
                              y: pd.DataFrame,
                              span: Optional[int] = 31,
                              ewm_lambda: float = 0.94,
                              is_x_correlated: bool = True,
                              mean_adj_type: MeanAdjType = MeanAdjType.NONE,
                              init_type: InitType = InitType.MEAN
                              ) -> LinearModel:

    """
    estimate time series ewm betas
    factor_returns is factors returns data: T*M
    asset_returns is asset returns data: T*N
    return type is {factor_returns.columns: factor of asset_returns T*N}

    is_independent specifies the asset_returns covar structure
    is_x_correlated:
    True = x covariance is orthogonal with ewm_covar being diagonal matrix
    False = x covariance is ewm covariance with ewm_covar being full matrix

    ewm betas = (ewm_covar[x, x]^{-1}) @ (ewm_covar[x, y])
    """
    if span is not None:
        ewm_lambda = 1.0 - 2.0/(span+1.0)
    if mean_adj_type != MeanAdjType.NONE:
        x = ewm.compute_rolling_mean_adj(data=x,
                                         mean_adj_type=mean_adj_type,
                                         ewm_lambda=ewm_lambda,
                                         init_type=init_type)

        y = ewm.compute_rolling_mean_adj(data=y,
                                         mean_adj_type=mean_adj_type,
                                         ewm_lambda=ewm_lambda,
                                         init_type=init_type)

    # compute list of betas using ewm numba recursion for cross product of x y and covariance of x
    # output is tensor of betas per date = [t, factors, assets]
    betas_ts = ewm.compute_ewm_xy_beta_tensor(x=x.to_numpy(),
                                              y=y.to_numpy(),
                                              ewm_lambda=ewm_lambda,
                                              is_x_correlated=is_x_correlated)
    # factor_loadings = {factor_id: pd.DataFrame(factor loadings)}
    loadings = dfo.np_txy_tensor_to_pd_dict(np_tensor_txy=betas_ts,
                                            dateindex=x.index,
                                            factor_names=x.columns.to_list(),
                                            asset_names=y.columns.to_list())

    ewm_lm = LinearModel(x=x, y=y, loadings=loadings)

    return ewm_lm


class UnitTests(Enum):
    MODEL_TEST = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.MODEL_TEST:

        from qis.test_data import load_etf_data
        prices = load_etf_data().dropna()

        returns = np.log(prices.divide(prices.shift(1)))

        # factors
        factors = ['SPY', 'TLT', 'GLD']
        factors = ['SPY']
        factor_returns = returns[factors]

        # assets
        is_check = False
        if is_check:
            asset_returns = returns[factors]
            asset_returns.columns = [f"{x.split('_')[0]}_asset" for x in factors]
        else:
            assets = ['QQQ', 'HYG']
            asset_returns = returns[assets]

        ewm_linear_model = estimate_ewm_linear_model(x=factor_returns,
                                                     y=asset_returns,
                                                     ewm_lambda=0.94,
                                                     is_x_correlated=True)

        ewm_linear_model.print()
        ewm_linear_model.plot_factor_loadings(factor='SPY')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MODEL_TEST

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Union
from enum import Enum

# qis
import qis.file_utils as fu
import qis


# strats
from qis.portfolio.strats.quant_strats_delta1 import simulate_vol_target_strats_range, simulate_trend_starts_range

FIG_SIZE = (8.3, 11.7)  # A4 for portrait

PERF_PARAMS = qis.PerfParams(freq='B')


def create_time_series_report(prices: Union[pd.Series, pd.DataFrame],
                              time_period: qis.TimePeriod,
                              spans: List[int] = (7, 14, 21, 30, 60, 130, 260, 520),
                              vol_span: int = 31,
                              vol_target: float = 0.15,
                              vol_af: float = 260
                              ) -> List[plt.Figure]:

    if isinstance(prices, pd.Series):
        strat_name = prices.name
    else:
        strat_name = ''

    # vol target returns
    vt_nav_weights, vt_navs = simulate_vol_target_strats_range(prices=prices, vol_spans=spans,
                                                               vol_target=vol_target, vol_af=vol_af)

    fig1 = plt.figure(figsize=FIG_SIZE, constrained_layout=True)
    nrows = 5
    gs = fig1.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
    axs1 = [fig1.add_subplot(gs[row, 0]) for row in range(nrows)]
    axs2 = [fig1.add_subplot(gs[row, 1]) for row in range(nrows)]
    fig1.suptitle(f"{strat_name} Volatility-target strats with vol_target={vol_target:0.0%}", fontweight="bold", fontsize=8, color='blue')

    global_kwargs = dict(fontsize=5, linewidth=0.5, first_color_fixed=True, framealpha=0.8,
                         digits_to_show=1, x_date_freq='A')
    scatter_kwargs = qis.update_kwargs(kwargs=global_kwargs,
                                       new_kwargs=dict(fontsize=5, var_format='{:.0%}', xvar_numticks=7,
                                                       markersize=1))

    plot_strategies_prices(nav_data=vt_navs,
                           nav_weights=vt_nav_weights,
                           time_period=time_period,
                           vol_span=vol_span,
                           vol_af=vol_af,
                           axs=axs1,
                           **global_kwargs)
    plot_strategies_returns_scatter(nav_data=vt_navs,
                                    time_period=time_period,
                                    axs=axs2,
                                    **scatter_kwargs)

    # tf returns
    tf_nav_weights, tf_navs, tf_signals = simulate_trend_starts_range(prices=prices, tf_spans=spans, vol_span=vol_span,
                                                                      vol_target=vol_target/2.0, vol_af=vol_af)

    fig2 = plt.figure(figsize=FIG_SIZE, constrained_layout=True)
    gs = fig2.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
    axs1 = [fig2.add_subplot(gs[row, 0]) for row in range(nrows)]
    axs2 = [fig2.add_subplot(gs[row, 1]) for row in range(nrows)]
    fig2.suptitle(f"{strat_name} Trend-following strats with vol_target={vol_target/2.0:0.0%}", fontweight="bold", fontsize=8, color='blue')

    plot_strategies_prices(nav_data=tf_navs,
                           nav_weights=tf_nav_weights,
                           time_period=time_period,
                           vol_span=vol_span,
                           vol_af=vol_af,
                           axs=axs1,
                           **global_kwargs)

    """
    figure_num += 1
    figure_caption = f"Figure {figure_num}. Pdf plot and qq-plot of returns on Trend-following Strategies by frequency"
    fig = plot_strategies_returns_pdf(nav_data=tf_navs,
                                      time_period=time_period,
                                      **global_kwargs)
    dut.add_fig_par(document, fig, figure_caption=figure_caption, figure_name=f"{file_name}_{figure_num}")
    """

    plot_strategies_returns_scatter(nav_data=tf_navs,
                                    time_period=time_period,
                                    axs=axs2,
                                    **scatter_kwargs)

    return [fig1, fig2]


def plot_strategies_prices(nav_data: pd.DataFrame,
                           nav_weights: pd.DataFrame,
                           time_period: qis.TimePeriod,
                           axs: List[plt.Subplot],
                           vol_span: int = 31,
                           vol_af: float = 260.0,
                           **kwargs
                           ) -> None:

    nav_returns = qis.to_returns(prices=nav_data)
    eod_ewm_vol = qis.compute_ewm_vol(data=nav_returns, span=vol_span, mean_adj_type=qis.MeanAdjType.NONE, af=vol_af)

    # trim plot data
    nav_data = time_period.locate(nav_data)
    nav_weights = time_period.locate(nav_weights)
    nav_returns = time_period.locate(nav_returns)
    eod_ewm_vol = time_period.locate(eod_ewm_vol)

    qis.plot_prices(prices=nav_data,
                    var_format='{:.2f}',
                    title='Logarithm of nav',
                    performance_label=qis.PerformanceLabel.WITH_SKEW,
                    is_log=True,
                    perf_params=PERF_PARAMS,
                    start_to_one=True,
                    ax=axs[0],
                    **kwargs)

    qis.plot_time_series(df=nav_weights,
                         var_format='{:.0%}',
                         title='Strategy weights',
                         legend_stats=qis.LegendStats.AVG_STD_LAST,
                         ax=axs[1],
                         **kwargs)

    turnover = nav_weights.diff(1).abs().rolling(int(vol_af)).sum()
    qis.plot_time_series(df=turnover,
                        var_format='{:.0%}',
                        title='1y rolling daily Turnover',
                        legend_stats=qis.LegendStats.AVG_STD_LAST,
                        ax=axs[2],
                        **kwargs)

    qis.plot_time_series(df=nav_returns,
                        var_format='{:.2%}',
                        title='Daily returns',
                        legend_stats=qis.LegendStats.AVG_STD_LAST,
                        ax=axs[3],
                        **kwargs)

    qis.plot_time_series(df=eod_ewm_vol,
                        var_format='{:.2%}',
                        title=f"Annualized EWMA-{vol_span} vol of daily returns",
                        legend_stats=qis.LegendStats.AVG_STD_LAST,
                        ax=axs[4],
                        **kwargs)


def plot_strategies_returns_pdf(nav_data: pd.DataFrame,
                                time_period: qis.TimePeriod,
                                **kwargs
                                ) -> plt.Figure:

    # trim plot data
    nav_data = time_period.locate(nav_data)

    freqs = {'Daily': 'B', 'Weekly': 'W-WED', 'Bi-Weekly': '2W-WED', 'Monthly': 'M'}

    numticks = 7
    major_ticks = np.linspace(-4.5, 4.5, numticks)
    pdf_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs={'fontsize': 7, 'var_format': '{:.0%}',
                                                              'xvar_numticks': numticks})

    qq_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs={'fontsize': 7, 'var_format': '{:.1f}',
                                                             'yvar_numticks': numticks,
                                                             'xvar_major_ticks': major_ticks})

    with sns.axes_style("darkgrid"):
        fig, axs = plt.subplots(len(freqs.keys()), 2, figsize=FIG_SIZE, tight_layout=True)
        for idx, (title, freq) in enumerate(freqs.items()):
            returns = qis.to_returns(prices=nav_data, freq=freq, drop_first=True)
            qis.plot_histogram(df=returns,
                               x_min_max_quantiles=(0.0001, 0.9999),
                               title=f"{title}",
                               add_data_std_pdf=True,
                               ax=axs[idx][0],
                               **pdf_kwargs)
            qis.plot_qq(df=returns,
                    title=f"{title}",
                    ax=axs[idx][1],
                    **qq_kwargs)
    return fig


def plot_strategies_returns_scatter(nav_data: pd.DataFrame,
                                    time_period: qis.TimePeriod,
                                    axs: List[plt.Subplot],
                                    **kwargs
                                    ) -> None:
    # trim plot data
    nav_data = time_period.locate(nav_data)
    freqs = {'Daily': ('B', 252), 'Weekly': ('W-WED', 52), 'Bi-Weekly': ('2W-WED', 26), 'Monthly': ('M', 12), 'Quarterly': ('Q', 4)}
    for ax, (title, freq) in zip(axs, freqs.items()):
        new_kwargs = dict(alpha_format='{0:+0.0%}',
                          beta_format='{:+0.1f}',
                          alpha_an_factor=freq[1],
                          framealpha=0.9)
        kwargs = qis.update_kwargs(kwargs, new_kwargs)
        qis.plot_returns_scatter(prices=nav_data,
                                 benchmark=nav_data.columns[0],
                                 ylabel=f"Strategy return",
                                 title=f"{title}",
                                 freq=freq[0],
                                 order=2,
                                 ci=95,
                                 ax=ax,
                                 **kwargs)


class UnitTests(Enum):
    BTC_SIMULATION = 1
    SPY_SIMULATION = 2


def run_unit_test(unit_test: UnitTests):

    import yfinance as yf

    if unit_test == UnitTests.BTC_SIMULATION:
        prices = yf.download(tickers=['BTC-USD'], start=None, end=None)['Adj Close'].rename('BTC').dropna()

        time_period = qis.TimePeriod('31Dec2015', '29Dec2022')
        figs = create_time_series_report(prices=prices, time_period=time_period, vol_target=0.5, vol_af=360)
        fu.save_figs_to_pdf(figs=figs, file_name='btc_analysis', orientation='landscape',
                            add_current_date=True, local_path=None)

    elif unit_test == UnitTests.SPY_SIMULATION:
        prices = yf.download(tickers=['SPY'], start=None, end=None)['Adj Close'].rename('SPY').dropna()
        time_period = qis.TimePeriod('31Dec1999', '29Dec2022')
        figs = create_time_series_report(prices=prices, time_period=time_period, vol_target=0.15, vol_af=260)
        fu.save_figs_to_pdf(figs=figs, file_name='spy_analysis', orientation='landscape',
                            add_current_date=True, local_path=None)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.SPY_SIMULATION

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import numpy as np
import pandas as pd
from typing import Union
from numba import njit
from statsmodels.tsa.stattools import pacf, acf

# qis
from qis.models.linear.ewm import MeanAdjType, compute_rolling_mean_adj


def estimate_path_acf(paths: Union[np.ndarray, pd.DataFrame],
                      nlags: int = 10,
                      is_pacf: bool = True
                      ):
    if isinstance(paths, pd.DataFrame):
        columns = paths.columns
        paths = paths.to_numpy()
    else:
        columns = None

    nb_path = paths.shape[1]
    acfs = np.zeros((nlags + 1, nb_path))
    for path in np.arange(nb_path):
        data = paths[:, path]
        data = data[np.isnan(data)==False]
        if len(data) > 2.0*nlags:
            if is_pacf:
                acfs[:, path] = pacf(data, nlags=nlags)
            else:
                acfs[:, path] = acf(data, nlags=nlags)
        else:
            acfs[:, path] = np.nan
    acfs = pd.DataFrame(acfs, columns=columns)
    m_acf = pd.Series(np.nanmean(acfs, axis=1), name='mean')
    std_acf = pd.Series(np.nanstd(acfs, axis=1), name='str')
    return acfs, m_acf, std_acf


def compute_autocorr(x: Union[np.ndarray, pd.Series],
                     num_lags: int = 1,
                     axis: int = None
                     ) -> np.ndarray:
    """
    partial correlation
    """
    if isinstance(x, pd.Series) or isinstance(x, pd.DataFrame):
        x = x.to_numpy()
    if len(x) == 1:
        corr = np.zeros(num_lags)
    else:
        corr = auto_corr(num_lags=num_lags)
    if num_lags == 1:
        return corr[0]
    else:
        return np.array(corr)


@njit
def auto_corr(a: np.ndarray,
              num_lags: int = 20
              ) -> np.ndarray:
    acorr = np.ones(num_lags)
    for idx in range(1, num_lags):
        acorr[idx] = np.corrcoef(a[idx:], a[:-idx], rowvar=False)[0][1]
    return acorr


@njit
def compute_ewm_autocovar(a: np.ndarray,
                           ewm_lambda: float = 0.94,
                           covar0: np.ndarray = None,
                           lag: int = 1,
                           aggregation_type: str = 'mean',
                           is_normalize: bool = True
                           ) -> (np.ndarray, np.ndarray):
    """
    x is T*N arrays
    """
    ewm_lambda_1 = 1.0 - ewm_lambda

    if a.ndim == 1:  # ndarry
        raise TypeError(f"time dimension must be higher than one")
    else:
        n = a.shape[1]  # array of ndarray
        num_off_diag = n * (n - 1)
        t = a.shape[0]

    if covar0 is None:
        covar = np.zeros((n, n))
        auto_covar = np.zeros((n, n))
    else:
        covar = covar0
        auto_covar = covar0

    trace_cov = np.zeros(t)
    trace_off = np.zeros(t)
    for idx in range(lag, t):  # row in x:
        x_t = a[idx]
        auto_covar = ewm_lambda_1 * np.outer(a[idx - lag], x_t) + ewm_lambda * auto_covar
        covar = ewm_lambda_1 * np.outer(x_t, x_t) + ewm_lambda * covar

        if is_normalize:
            auto_covar_t = auto_covar / covar
        else:
            auto_covar_t = auto_covar

        if aggregation_type == 'mean':
            trace_cov[idx] = np.nansum(np.diag(auto_covar_t)) / n
            trace_off[idx] = (np.nansum(auto_covar_t) - np.nansum(np.diag(auto_covar_t)))/num_off_diag

        elif aggregation_type == 'median':
            trace_cov[idx] = np.nanmedian(np.diag(auto_covar_t))
            trace_off[idx] = np.nanmedian(auto_covar_t)

    return trace_cov, trace_off


def compute_auto_corr(data: pd.DataFrame,
                      num_lags: int = 20
                      ) -> pd.DataFrame:
    """
    compute auto correlation columns wise
    """
    acorrs = []
    for column in data:
        acorrs.append(auto_corr(a=data[column].dropna().to_numpy(), num_lags=num_lags))

    data = pd.DataFrame(data=np.column_stack(acorrs),
                        index=range(0, num_lags),
                        columns=data.columns)
    return data


def compute_dynamic_auto_corr(data: pd.DataFrame,
                              ewm_lambda: float = 0.94,
                              mean_adj_type: MeanAdjType = MeanAdjType.EWMA,
                              lag: int = 1,
                              aggregation_type: str = 'mean',
                              is_normalize: bool = True
                              ) -> pd.DataFrame:
    """
    compute auto correlation columns wise
    """
    if len(data.index) == 1:
        raise TypeError('data must be time series')

    data = data.fillna(method='ffill').dropna()
    x = compute_rolling_mean_adj(data=data.to_numpy(),
                                 mean_adj_type=mean_adj_type,
                                 ewm_lambda=ewm_lambda)

    trace_cov, trace_corr = compute_ewm_autocovar(a=x,
                                                   ewm_lambda=ewm_lambda,
                                                   lag=lag,
                                                   aggregation_type=aggregation_type,
                                                   is_normalize=is_normalize)

    data = pd.DataFrame(data=np.column_stack((trace_cov, trace_corr)),
                        index=data.index,
                        columns=['diagonal', 'off-diag'])
    return data



"""
illustrations for paper
"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import qis
import yfinance as yf


FIG_SIZE21 = (15, 12)
FIG_SIZE12 = (15, 7)
LOCAL_PATH = "..//draft//figures//"

MOVE_NAME = 'Move volatility index'
fixed_tenor = '10y'
RVOL = 'Realized 10y rate Vol'

move = yf.download(tickers=['^MOVE'], start=None, end=None)['Adj Close'].rename(MOVE_NAME)

tickers = {'^IRX': '3m', '^FVX': '5y', '^TNX': '10y', '^TYX': '30y'}
tenors = list(tickers.values())

rates = yf.download(tickers=list(tickers.keys()), start=None, end=None)['Adj Close'].rename(tickers, axis=1)

rates = rates.reindex(index=move.index, method='ffill')*100.0

move = move.loc[:'2022']
rates = rates.loc[:'2022', :]

freq = 'W-MON'
move_change = move.resample(freq).last().diff(1).loc[:'2022']
rates_change = rates.resample(freq).last().diff(1).loc[:'2022', :]
real_vol = np.sqrt(52.0*qis.compute_ewm(data=np.square(rates_change[fixed_tenor]), span=26)).rename(RVOL)
real_vol_change = real_vol.diff(1)

joint = pd.concat([rates, move, real_vol], axis=1).iloc[10:, :]
joint_change = pd.concat([rates_change, move_change, real_vol_change], axis=1).dropna()

vol_beta = qis.compute_one_factor_ewm_betas(x=joint_change[fixed_tenor], y=joint_change[MOVE_NAME].to_frame(), span=26
                                            ).iloc[:, 0].rename('Realized Volatility Beta')
vol_vol = np.sqrt(260.0*qis.compute_ewm(data=np.square(joint_change[MOVE_NAME]), span=26, init_type=qis.InitType.MEAN)
                  ).rename('Realized Volatility-of-Volatility')

hue = 'year'
fixed_years = [2001, 2007, 2010, 2017, 2020, 2022]
colors = ['orchid', 'green', 'grey', 'olive', 'red']

with sns.axes_style('darkgrid'):

    fig, axs = plt.subplots(2, 2, figsize=(8, 8), tight_layout=True)
    qis.set_suptitle(fig=fig, title='Tenor rates vs Move Index')
    for idx, tenor in enumerate(tenors):
        df = qis.add_hue_fixed_years(df=joint[[tenor, MOVE_NAME]].copy(), hue=hue, fixed_years=fixed_years)
        qis.plot_scatter(df=df,
                         hue=hue,
                         xvar_format='{:.0%}',
                         yvar_format='{:.0%}',
                         order=2,
                         ci=None,
                         title=f"{tenor}",
                         ax=axs[idx // 2][idx % 2])

    fig, axs = plt.subplots(2, 2, figsize=(8, 8), tight_layout=True)
    qis.set_suptitle(fig=fig, title='Tenor rates change vs Change in Move Index')
    for idx, tenor in enumerate(tenors):
        df = qis.add_hue_fixed_years(df=joint_change[[tenor, MOVE_NAME]].copy(), hue=hue, fixed_years=fixed_years)
        qis.plot_scatter(df=df,
                         hue=hue,
                         xvar_format='{:.2%}',
                         yvar_format='{:.2%}',
                         order=2,
                         ci=None,
                         title=f"{tenor}",
                         ax=axs[idx//2][idx%2])


    kwargs = dict(fontsize=14, legend_loc='upper center', linewidth=1.5, framealpha=0.75, x_date_freq='A')

    def add_shadows(ax: plt.Subplot):
        df = qis.add_hue_fixed_years(df=rates.copy(), hue=hue, fixed_years=fixed_years)
        for idx, (key, df_) in enumerate(df.groupby(hue)):
            ax.axvspan(xmin=df_.index[0], xmax=df_.index[-1], alpha=0.2, color=colors[idx], lw=0)
        ax.set_xlim([rates.index[0], rates.index[-1]])


    #################################
    # time series figures for article
    ################################
    fig, axs = plt.subplots(2, 1, figsize=FIG_SIZE21, tight_layout=True)
    qis.plot_time_series_2ax(df1=joint[[fixed_tenor]], df2=joint[[MOVE_NAME, RVOL]],
                             var_format='{:,.0f}',
                             var_format_yax2='{:,.0f}',
                             legend_stats=qis.LegendStats.FIRST_AVG_LAST,
                             legend_stats2=qis.LegendStats.FIRST_AVG_LAST,
                             y_limits=(0.0, None), y_limits_ax2=(0.0, None),
                             title=f"(A) Dynamics of {fixed_tenor} rate and {MOVE_NAME} in bp",
                             x_rotation=90,
                             ax=axs[0],
                             **kwargs)
    add_shadows(ax=axs[0])

    # beta and vol vol
    new_kwargs = dict(yvar_major_ticks1=np.linspace(-0.8, 0.8, 5), yvar_major_ticks2=np.linspace(0, 500.0, 6))
    qis.plot_time_series_2ax(df1=vol_beta, df2=vol_vol,
                             var_format='{:,.2f}',
                             var_format_yax2='{:,.0f}',
                             legend_stats=qis.LegendStats.FIRST_AVG_LAST,
                             legend_stats2=qis.LegendStats.FIRST_AVG_LAST,
                             #y_limits=(0.0, None), y_limits_ax2=(0.0, None),
                             title=f"(B) Realized rates beta and volatility-of-volatility of {MOVE_NAME}",
                             x_rotation=90,
                             ax=axs[1],
                             **qis.update_kwargs(kwargs, new_kwargs))
    add_shadows(ax=axs[1])
    qis.save_fig(fig=fig, file_name='timeseries', local_path=LOCAL_PATH)

    # pdf of vols
    fig, axs = plt.subplots(1, 2, figsize=FIG_SIZE12, tight_layout=True)
    qis.plot_histogram(df=joint[[MOVE_NAME, RVOL]],
                       desc_table_type=qis.DescTableType.WITH_KURTOSIS,
                       title=f"Vols",
                       #x_min_max_quantiles=(0.01, 0.99),
                       ax=axs[0],
                       **kwargs)
    qis.plot_histogram(df=np.log(joint[[MOVE_NAME, RVOL]]),
                       desc_table_type=qis.DescTableType.WITH_KURTOSIS,
                       title=f"Log Vols",
                       #x_min_max_quantiles=(0.01, 0.99),
                       ax=axs[1],
                       **kwargs)

    # qqplots
    fig, axs = plt.subplots(1, 2, figsize=FIG_SIZE12, tight_layout=True)
    qis.plot_qq(df=joint[[MOVE_NAME, RVOL]],
                title=f"Vols",
                #x_min_max_quantiles=(0.01, 0.99),
                ax=axs[0],
                **kwargs)
    qis.plot_qq(df=np.log(joint[[MOVE_NAME, RVOL]]),
                title=f"Log Vols",
                #x_min_max_quantiles=(0.01, 0.99),
                ax=axs[1],
                **kwargs)

    # move and real vols vs rate
    fig, axs = plt.subplots(1, 2, figsize=FIG_SIZE12, tight_layout=True)
    df = qis.add_hue_fixed_years(df=joint[[fixed_tenor, MOVE_NAME]].copy(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xvar_format='{:,.0f}',
                     yvar_format='{:,.0f}',
                     hue=hue,
                     order=1,
                     fit_intercept=True,
                     add_hue_model_label=True,
                     full_sample_order=0,
                     markersize=12,
                     colors=colors,
                     ci=None,
                     title=f"(A) {MOVE_NAME} vs {fixed_tenor} rate",
                     ax=axs[0],
                     **kwargs)

    df = qis.add_hue_fixed_years(df=joint[[fixed_tenor, RVOL]].copy(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xvar_format='{:,.0f}',
                     yvar_format='{:,.0f}',
                     hue=hue,
                     order=1,
                     fit_intercept=True,
                     add_hue_model_label=True,
                     full_sample_order=0,
                     markersize=12,
                     colors=colors,
                     ci=None,
                     title=f"(B) Realized volatility vs {fixed_tenor} rate",
                     ax=axs[1],
                     **kwargs)

    # change in move and real vol vs change in rate
    fig, axs = plt.subplots(1, 2, figsize=FIG_SIZE12, tight_layout=True)
    df = qis.add_hue_fixed_years(df=joint_change[[fixed_tenor, MOVE_NAME]].copy(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xlabel=f"Change in {fixed_tenor} rate",
                     ylabel=f"Change in {MOVE_NAME}",
                     xvar_format='{:,.0f}',
                     yvar_format='{:,.0f}',
                     hue=hue,
                     order=2,
                     full_sample_order=2,
                     markersize=12,
                     colors=colors,
                     ci=None,
                     fit_intercept=True,
                     add_universe_model_label=False,
                     title=f"(A) Change in {MOVE_NAME} vs change in {fixed_tenor} rate",
                     ax=axs[0],
                     **kwargs)
    df = qis.add_hue_fixed_years(df=joint_change[[fixed_tenor, RVOL]].copy(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xlabel=f"Change in {fixed_tenor} rate",
                     ylabel=f"Change in realized volatility",
                     xvar_format='{:.0%}',
                     yvar_format='{:.0%}',
                     hue=hue,
                     order=2,
                     full_sample_order=2,
                     markersize=12,
                     colors=colors,
                     ci=None,
                     fit_intercept=True,
                     add_universe_model_label=False,
                     title=f"(B) Change in realized volatility vs change in {fixed_tenor} rate",
                     ax=axs[1],
                     **kwargs)

    #################################
    # move vs rate and move changes vs rate figures for article
    ################################
    fig, axs = plt.subplots(1, 2, figsize=FIG_SIZE12, tight_layout=True)
    df = qis.add_hue_fixed_years(df=joint[[fixed_tenor, MOVE_NAME]].copy(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xvar_format='{:,.0f}',
                     yvar_format='{:,.0f}',
                     hue=hue,
                     order=1,
                     fit_intercept=False,
                     add_hue_model_label=True,
                     full_sample_order=0,
                     markersize=12,
                     colors=colors,
                     ci=None,
                     title=f"(A) {MOVE_NAME} vs {fixed_tenor} rate",
                     ax=axs[0],
                     **kwargs)

    df = qis.add_hue_fixed_years(df=joint_change[[fixed_tenor, MOVE_NAME]].copy(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xlabel=f"Change in {fixed_tenor} rate",
                     ylabel=f"Change in {MOVE_NAME}",
                     xvar_format='{:,.0f}',
                     yvar_format='{:,.0f}',
                     hue=hue,
                     order=2,
                     full_sample_order=2,
                     markersize=12,
                     colors=colors,
                     ci=None,
                     fit_intercept=True,
                     add_universe_model_label=False,
                     title=f"(B) Change in {MOVE_NAME} vs change in {fixed_tenor} rate",
                     ax=axs[1],
                     **kwargs)

    qis.save_fig(fig=fig, file_name='a_move_b_change_vs_rate', local_path=LOCAL_PATH)

    #################################
    # realized vol-vol and vol-beta vs move index for article
    ################################
    fig, axs = plt.subplots(1, 2, figsize=FIG_SIZE12, tight_layout=True)
    df = qis.add_hue_fixed_years(df=pd.concat([move, vol_vol], axis=1).dropna(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xvar_format='{:,.0f}',
                     yvar_format='{:,.0f}',
                     hue=hue,
                     order=1,
                     full_sample_order=0,
                     fit_intercept=False,
                     markersize=12,
                     colors=colors,
                     ci=None,
                     title=f"(A) Realized volatility-of-volatility vs {MOVE_NAME}",
                     ax=axs[0],
                     **kwargs)
    # volbeta vs move
    df = qis.add_hue_fixed_years(df=pd.concat([move, vol_beta], axis=1).dropna(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xvar_format='{:.0f}',
                     yvar_format='{:.2f}',
                     beta_format='{0:+0.4f}',
                     hue=hue,
                     order=1,
                     full_sample_order=0,
                     fit_intercept=False,
                     markersize=12,
                     colors=colors,
                     ci=None,
                     title=f"(B) Realized rates beta vs {MOVE_NAME}",
                     ax=axs[1],
                     **kwargs)
    qis.save_fig(fig=fig, file_name='move_volvol', local_path=LOCAL_PATH)

    # realized volvol and beta vs rate
    fig, axs = plt.subplots(1, 2, figsize=FIG_SIZE12, tight_layout=True)
    df = qis.add_hue_fixed_years(df=pd.concat([rates[fixed_tenor], vol_vol], axis=1).dropna(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xvar_format='{:,.0f}',
                     yvar_format='{:,.0f}',
                     hue=hue,
                     order=1,
                     full_sample_order=0,
                     fit_intercept=False,
                     markersize=12,
                     colors=colors,
                     ci=95,
                     title=f"Realized Volvol vs {fixed_tenor} rate",
                     ax=axs[0],
                     **kwargs)
    df = qis.add_hue_fixed_years(df=pd.concat([rates[fixed_tenor], vol_beta], axis=1).dropna(), hue=hue, fixed_years=fixed_years)
    qis.plot_scatter(df=df,
                     xvar_format='{:,.0f}',
                     yvar_format='{:,.2f}',
                     hue=hue,
                     order=1,
                     full_sample_order=0,
                     fit_intercept=False,
                     markersize=12,
                     colors=colors,
                     ci=95,
                     title=f"Realized beta vs {fixed_tenor} rate",
                     ax=axs[1],
                     **kwargs)

plt.show()



"""
plot tables for regime classification
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Optional
from enum import Enum

# qis
import qis.utils.df_cut as dfc
import qis.utils.df_str as dfs
import qis.plots.table as ptb


def get_quantile_class_table(data: pd.DataFrame,
                             x_column: str,
                             y_column: Optional[str] = None,
                             num_buckets: int = 4,
                             y_data_lag: Optional[int] = None,
                             hue_name: str = 'hue',
                             xvar_format: str = '{:.2f}'
                             ) -> pd.DataFrame:
    scatter_data, _ = dfc.add_quantile_classification(df=data, x_column=x_column, hue_name=hue_name,
                                                      num_buckets=num_buckets,
                                                      bins=None,
                                                      xvar_format=xvar_format)
    if y_column is None:
        y_column = x_column

    if y_data_lag is not None:
        scatter_data[y_column] = scatter_data[y_column].shift(y_data_lag)
        # scatter_data = scatter_data.dropna()

    stats = scatter_data[[y_column, hue_name]].groupby(hue_name, sort=False).agg(['count', 'mean', 'std'])
    stats.columns = stats.columns.get_level_values(1)
    stats.insert(loc=0, column='freq', value=stats['count'].to_numpy() / np.sum(stats['count'].to_numpy()))
    return stats


def plot_quantile_class_table(data: pd.DataFrame,
                              x_column: str,
                              y_column: Optional[str] = None,
                              num_buckets: int = 4,
                              hue_name: str = 'hue',
                              var_format: str = '{:.2%}',
                              xvar_format: str = '{:.2f}',
                              y_data_lag: Optional[int] = None,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:
    stats = get_quantile_class_table(data=data, x_column=x_column, y_column=y_column,
                                     num_buckets=num_buckets,
                                     y_data_lag=y_data_lag,
                                     hue_name=hue_name,
                                     xvar_format=xvar_format)

    # stats is freq, count, mean and std
    stats = dfs.df_to_str(stats, var_formats=['{:.0%}', '{:.0f}', var_format, var_format])

    ptb.plot_df_table(df=stats,
                      index_column_name=hue_name,
                      ax=ax,
                      **kwargs)


class UnitTests(Enum):
    QUANTILE_CLASS_TABLE = 1


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()
    returns = prices.asfreq('Q', method='ffill').pct_change().dropna()

    if unit_test == UnitTests.QUANTILE_CLASS_TABLE:
        plot_quantile_class_table(data=returns, x_column='SPY', num_buckets=4, hue_name='quantile regime')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.QUANTILE_CLASS_TABLE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)


import yaml
from pathlib import Path
from typing import Dict


def get_paths() -> Dict[str, str]:
    """
    read path specs in settings.yaml
    """
    full_file_path = Path(__file__).parent.joinpath('settings.yaml')
    with open(full_file_path) as settings:
        settings_data = yaml.load(settings, Loader=yaml.Loader)
    return settings_data


def get_resource_path() -> str:
    """
    read path specs in settings.yaml
    """
    full_file_path = Path(__file__).parent.joinpath('settings.yaml')
    with open(full_file_path) as settings:
        settings_data = yaml.load(settings, Loader=yaml.Loader)
    return settings_data['RESOURCE_PATH']


def get_output_path() -> str:
    """
    read path specs in settings.yaml
    """
    full_file_path = Path(__file__).parent.joinpath('settings.yaml')
    with open(full_file_path) as settings:
        settings_data = yaml.load(settings, Loader=yaml.Loader)
    return settings_data['OUTPUT_PATH']




import qis.local_path

from qis.file_utils import (
    delocalize_df,
    save_figs_to_pdf,
    get_all_folder_files,
    get_local_file_path,
    get_output_path,
    get_param_file_path,
    get_paths,
    get_pdf_path,
    get_resource_path,
    join_file_name_parts,
    load_df_dict_from_csv,
    load_df_dict_from_excel,
    load_df_dict_from_feather,
    load_df_dict_from_parquet,
    load_df_dict_from_sql,
    load_df_from_csv,
    load_df_from_excel,
    load_df_from_feather,
    load_df_from_parquet,
    save_df_dict_to_csv,
    save_df_dict_to_excel,
    save_df_dict_to_feather,
    save_df_dict_to_parquet,
    save_df_dict_to_sql,
    save_df_to_csv,
    save_df_to_excel,
    save_df_to_feather,
    save_df_to_parquet,
    save_fig,
    save_figs,
    timer
)

from qis.utils.__init__ import *

from qis.perfstats.__init__ import *

from qis.plots.__init__ import *

from qis.models.__init__ import *

from qis.portfolio.__init__ import *



"""
core implementation of baktest report focus on strategy and comparison vs benchmark stategy
"""

# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from dataclasses import dataclass
from typing import List, Optional

# qis
import qis.plots.derived.regime_data
import qis.utils.dates as da
import qis.utils.struct_ops as sop
import qis.utils.df_str as dfs
import qis.utils.df_groups as dfg
import qis.perfstats.returns as ret
import qis.perfstats.perf_stats as rpt
import qis.plots.derived.drawdowns as cdr
from qis.perfstats.config import PerfParams, PerfStat, RegimeData
import qis.perfstats.regime_classifier as rcl
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs

# plots
import qis.plots.table as ptb
import qis.plots.time_series as pts
import qis.plots.derived.prices as ppd
import qis.plots.derived.returns_heatmap as rhe
import qis.plots.derived.perf_table as ppt
import qis.plots.derived.returns_scatter as prs
import qis.models.linear.plot_correlations as pco
from qis.plots.bars import plot_bars

# internal
from qis.portfolio.portfolio_data import PortfolioData, AttributionMetric


PERF_PARAMS = PerfParams(freq='W-WED')
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='M')


@dataclass
class MultiPortfolioData:
    """
    data structure to unify multi portfolio reporting
    """
    portfolio_datas: List[PortfolioData]
    benchmark_prices: pd.DataFrame = None

    def __post_init__(self):
        self.set_navs(freq=None)  # default frequency is freq of backtests, can be non for strats at different freqs

    def set_navs(self, freq: Optional[str] = None):
        navs = []
        for portfolio in self.portfolio_datas:
            navs.append(portfolio.get_portfolio_nav())
        self.navs = pd.concat(navs, axis=1)

        if freq is not None:
            self.navs = self.navs.asfreq(freq=freq, method='ffill')

        if self.benchmark_prices is not None:
            self.benchmark_prices = self.benchmark_prices.reindex(index=self.navs.index, method='ffill')

    def _set_benchmark_prices(self, benchmark_prices: pd.DataFrame) -> None:
        self.benchmark_prices = benchmark_prices
        self.set_navs(freq=None)

    """
    data get methods
    """
    def get_navs(self,
                 benchmark: str = None,
                 time_period: da.TimePeriod = None
                 ) -> pd.DataFrame:
        """
        get portfolio navs
        """
        navs = self.navs
        if benchmark is not None:
            navs = pd.concat([self.benchmark_prices[benchmark], navs], axis=1).fillna(method='ffill')
        if time_period is not None:
            navs = time_period.locate(navs)
        return navs

    def get_benchmark_price(self,
                            benchmark: str,
                            time_period: da.TimePeriod = None
                            ) -> pd.Series:
        price = self.benchmark_prices[benchmark]
        if time_period is not None:
            price = time_period.locate(price)
        return price

    def get_ac_navs(self,
                    portfolio_idx: int = 0,
                    benchmark: str = None,
                    time_period: da.TimePeriod = None
                    ) -> pd.DataFrame:
        prices = self.portfolio_datas[portfolio_idx].get_ac_navs(time_period=time_period)
        if benchmark is not None:
            benchmark_price = self.get_benchmark_price(benchmark=self.benchmark_prices.columns[0],
                                                       time_period=time_period)
            prices = pd.concat([prices, benchmark_price], axis=1)
        return prices

    """
    plot methods
    """
    def add_regime_shadows(self, ax: plt.Subplot, benchmark: str, index: pd.Index = None,
                           regime_params: BenchmarkReturnsQuantileRegimeSpecs=REGIME_PARAMS
                           ) -> None:
        pivot_prices = self.benchmark_prices[benchmark]
        if index is not None:
            pivot_prices = pivot_prices.reindex(index=index, method='ffill')
        qis.plots.derived.regime_data.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)

    def plot_nav(self,
                 time_period: da.TimePeriod = None,
                 benchmark: str = None,
                 perf_params: PerfParams = PERF_PARAMS,
                 regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                 ax: plt.Subplot = None,
                 **kwargs) -> None:

        if ax is None:
            fig, ax = plt.subplots()

        prices = self.get_navs(time_period=time_period)
        ppd.plot_prices(prices=prices,
                        perf_params=perf_params,
                        ax=ax,
                        **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=prices.index, regime_params=regime_params)

    def plot_periodic_returns(self,
                              time_period: da.TimePeriod = None,
                              heatmap_freq: str = 'A',
                              date_format: str = '%Y',
                              transpose: bool = True,
                              title: str = None,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:
        prices = self.get_navs(time_period=time_period)
        rhe.plot_periodic_returns_table(prices=prices,
                                                    freq=heatmap_freq,
                                                    ax=ax,
                                                    title=title,
                                                    date_format=date_format,
                                                    transpose=transpose,
                                                    **kwargs)

    def plot_performance_bars(self,
                              time_period: da.TimePeriod = None,
                              perf_column: PerfStat = PerfStat.SHARPE,
                              perf_params: PerfParams = PERF_PARAMS,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:

        prices = self.get_navs(time_period=time_period)
        ppt.plot_ra_perf_bars(prices=prices,
                              perf_column=perf_column,
                              perf_params=perf_params,
                              title=f"{perf_column.to_str()}: {da.get_time_period(prices).to_str()}",
                              ax=ax,
                              **kwargs)

    def plot_corr_table(self,
                        time_period: da.TimePeriod = None,
                        freq: str = 'W-WED',
                        ax: plt.Subplot = None,
                        **kwargs) -> None:
        prices = self.get_navs(time_period=time_period)
        if len(prices.columns) > 1:
            pco.plot_corr_table(prices=prices,
                                x_rotation=90,
                                freq=freq,
                                title=f'Correlation of {freq} returns',
                                ax=ax,
                                **kwargs)

    def plot_drawdowns(self,
                       time_period: da.TimePeriod = None,
                       regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                       benchmark: str = None,
                       ax: plt.Subplot = None,
                       **kwargs) -> None:
        if len(self.portfolio_datas) == 1 and benchmark is not None:
            prices = self.get_navs(time_period=time_period, benchmark=benchmark)
        else:
            prices = self.get_navs(time_period=time_period)
        cdr.plot_drawdown(prices=prices, ax=ax, **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=prices.index, regime_params=regime_params)

    def plot_rolling_time_under_water(self,
                                      time_period: da.TimePeriod = None,
                                      regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                      benchmark: str = None,
                                      ax: plt.Subplot = None,
                                      **kwargs) -> None:
        if len(self.portfolio_datas) == 1 and benchmark is not None:
            prices = self.get_navs(time_period=time_period, benchmark=benchmark)
        else:
            prices = self.get_navs(time_period=time_period)
        cdr.plot_rolling_time_under_water(prices=prices, ax=ax, **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=prices.index, regime_params=regime_params)

    def plot_ra_perf_table(self,
                           time_period: da.TimePeriod = None,
                           perf_params: PerfParams = PERF_PARAMS,
                           ax: plt.Subplot = None,
                           **kwargs) -> None:
        benchmark = self.benchmark_prices.columns[0]
        prices = self.get_navs(benchmark=benchmark, time_period=time_period)
        ppt.plot_ra_perf_table_benchmark(prices=prices,
                                         benchmark=benchmark,
                                         perf_params=perf_params,
                                         perf_columns=rpt.BENCHMARK_TABLE_COLUMNS,
                                         title=f"RA performance table: {da.get_time_period(prices).to_str()}",
                                         rotation_for_columns_headers=0,
                                         ax=ax,
                                         **kwargs)

    def plot_ac_ra_perf_table(self,
                              benchmark_price: pd.Series,
                              time_period: da.TimePeriod = None,
                              perf_params: PerfParams = PERF_PARAMS,
                              ax: plt.Subplot = None,
                              **kwargs) -> None:
        strategy_prices = []
        ac_prices = []
        rows_edge_lines = [len(self.portfolio_datas)]
        for portfolio in self.portfolio_datas:
            portfolio_name = str(portfolio.nav.name)
            prices_ = portfolio.get_ac_navs(time_period=time_period)
            strategy_prices.append(prices_[portfolio_name])
            ac_prices_ = prices_.drop(portfolio_name, axis=1)
            ac_prices_.columns = [f"{portfolio_name}-{x}" for x in ac_prices_.columns]
            ac_prices.append(ac_prices_)
            rows_edge_lines.append(sum(rows_edge_lines)+len(ac_prices_.columns))
        strategy_prices = pd.concat(strategy_prices, axis=1)
        ac_prices = pd.concat(ac_prices, axis=1)
        prices = pd.concat([benchmark_price.reindex(index=strategy_prices.index, method='ffill'), strategy_prices, ac_prices], axis=1)
        ppt.plot_ra_perf_table_benchmark(prices=prices,
                                         benchmark=str(benchmark_price.name),
                                         perf_params=perf_params,
                                         perf_columns=rpt.BENCHMARK_TABLE_COLUMNS,
                                         drop_benchmark=True,
                                         rows_edge_lines=rows_edge_lines,
                                         title=f"RA performance table by Asset Group: {da.get_time_period(prices).to_str()}",
                                         rotation_for_columns_headers=0,
                                         row_height=0.5,
                                         ax=ax,
                                         **kwargs)

    def plot_exposures(self,
                       benchmark: str = None,
                       regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                       time_period: da.TimePeriod = None,
                       var_format: str = '{:.0%}',
                       ax: plt.Subplot = None,
                       **kwargs) -> None:
        exposures = []
        for portfolio in self.portfolio_datas:
            exposures.append(portfolio.get_exposures(time_period=time_period).sum(axis=1).rename(portfolio.nav.name))
        exposures = pd.concat(exposures, axis=1)
        pts.plot_time_series(df=exposures,
                             var_format=var_format,
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title='Portfolio net exposures',
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=exposures.index, regime_params=regime_params)

    def plot_instrument_pnl_diff(self,
                                 portfolio_idx1: int = 0,
                                 portfolio_idx2: int = 1,
                                 is_grouped: bool = True,
                                 benchmark: str = None,
                                 regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                 time_period: da.TimePeriod = None,
                                 var_format: str = '{:.0%}',
                                 ax: plt.Subplot = None,
                                 **kwargs) -> None:
        pnl_inst1 = self.portfolio_datas[portfolio_idx1].instrument_pnl
        pnl_inst2 = self.portfolio_datas[portfolio_idx2].instrument_pnl
        df1_, df2_ = pnl_inst1.align(other=pnl_inst2, join='outer', axis=None)
        df1_, df2_ = df1_.fillna(0.0), df2_.fillna(0.0)
        diff = df1_.subtract(df2_)

        if is_grouped:
            diff = dfg.agg_df_by_groups_ax1(diff,
                                            group_data=self.portfolio_datas[portfolio_idx1].group_data,
                                            agg_func=np.nansum,
                                            total_column=f"{self.portfolio_datas[portfolio_idx1].nav.name}-{self.portfolio_datas[portfolio_idx2].nav.name}",
                                            group_order=self.portfolio_datas[portfolio_idx1].group_order)
        if time_period is not None:
            diff = time_period.locate(diff)
        diff = diff.cumsum(axis=0)

        pts.plot_time_series(df=diff,
                             var_format=var_format,
                             legend_stats=pts.LegendStats.LAST,
                             title=f"Cumulative p&l diff {self.portfolio_datas[portfolio_idx1].nav.name}-{self.portfolio_datas[portfolio_idx2].nav.name}",
                             ax=ax,
                             **sop.update_kwargs(kwargs, dict(legend_loc='lower left')))
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=diff.index, regime_params=regime_params)

    def plot_exposures_diff(self,
                            portfolio_idx1: int = 0,
                            portfolio_idx2: int = 1,
                            benchmark: str = None,
                            regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                            time_period: da.TimePeriod = None,
                            var_format: str = '{:.0%}',
                            ax: plt.Subplot = None,
                            **kwargs) -> None:
        exposures1 = self.portfolio_datas[portfolio_idx1].get_exposures(is_grouped=True, time_period=time_period, add_total=False)
        exposures2 = self.portfolio_datas[portfolio_idx2].get_exposures(is_grouped=True, time_period=time_period, add_total=False)
        diff = exposures1.subtract(exposures2)
        pts.plot_time_series(df=diff,
                             var_format=var_format,
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title=f"Net exposure diff {self.portfolio_datas[portfolio_idx1].nav.name}-{self.portfolio_datas[portfolio_idx2].nav.name}",
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=diff.index, regime_params=regime_params)

    def plot_turnover(self,
                      roll_period: int = 260,
                      benchmark: str = None,
                      time_period: da.TimePeriod = None,
                      regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                      var_format: str = '{:.0%}',
                      ax: plt.Subplot = None,
                      **kwargs) -> None:

        turnover = []
        for portfolio in self.portfolio_datas:
            turnover.append(portfolio.get_turnover(roll_period=None, is_agg=True).rename(portfolio.nav.name))
        turnover = pd.concat(turnover, axis=1)
        if roll_period is not None:
            turnover = turnover.rolling(roll_period).sum()
        if time_period is not None:
            turnover = time_period.locate(turnover)

        pts.plot_time_series(df=turnover,
                             var_format=var_format,
                             y_limits=(0.0, None),
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title='Annualized daily Turnover',
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=turnover.index, regime_params=regime_params)

    def plot_costs(self,
                   roll_period: int = 260,
                   benchmark: str = None,
                   time_period: da.TimePeriod = None,
                   regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                   var_format: str = '{:.2%}',
                   ax: plt.Subplot = None,
                   **kwargs) -> None:
        costs = []
        for portfolio in self.portfolio_datas:
            costs.append(portfolio.get_costs(roll_period=None, is_agg=True).rename(portfolio.nav.name))
        costs = pd.concat(costs, axis=1)
        if roll_period is not None:
            costs = costs.rolling(roll_period).sum()
        if time_period is not None:
            costs = time_period.locate(costs)
        pts.plot_time_series(df=costs,
                             var_format=var_format,
                             y_limits=(0.0, None),
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title='Annualized daily Costs %',
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=costs.index, regime_params=regime_params)

    def plot_factor_betas(self,
                          benchmark_prices: pd.DataFrame,
                          benchmark: str = None,
                          time_period: da.TimePeriod = None,
                          regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                          var_format: str = '{:,.2f}',
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> None:
        factor_exposures = []
        for portfolio in self.portfolio_datas:
            factor_exposure = portfolio.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                          time_period=time_period)
            factor_exposure.columns = [f"{portfolio.nav.name}-{x} beta" for x in factor_exposure.columns]
            factor_exposures.append(factor_exposure)
        factor_exposures = pd.concat(factor_exposures, axis=1)
        pts.plot_time_series(df=factor_exposures,
                             var_format=var_format,
                             legend_stats=pts.LegendStats.AVG_NONNAN_LAST,
                             title='Factor betas',
                             ax=ax,
                             **kwargs)
        if benchmark is not None:
            self.add_regime_shadows(ax=ax, benchmark=benchmark, index=factor_exposures.index, regime_params=regime_params)

    def plot_nav_with_dd(self,
                         time_period: da.TimePeriod = None,
                         perf_params: PerfParams = PERF_PARAMS,
                         axs: List[plt.Subplot] = None,
                         **kwargs
                         ) -> None:
        prices = self.get_navs(time_period=time_period)
        if self.benchmark_prices is not None:
            regime_benchmark_str = self.benchmark_prices.columns[0]
        else:
            regime_benchmark_str = None
        ppd.plot_prices_with_dd(prices=prices,
                                perf_params=perf_params,
                                regime_benchmark_str=regime_benchmark_str,
                                axs=axs,
                                **kwargs)

    def plot_returns_scatter(self,
                             benchmark: str,
                             time_period: da.TimePeriod = None,
                             freq: str = 'Q',
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        prices = self.get_navs(benchmark=benchmark, time_period=time_period)
        local_kwargs = sop.update_kwargs(kwargs=kwargs,
                                         new_kwargs={'weight': 'bold',
                                                     #'alpha_an_factor': 52.0,
                                                     'x_rotation': 0,
                                                     'first_color_fixed': False,
                                                     'ci': None})
        prs.plot_returns_scatter(prices=prices,
                                 benchmark=benchmark,
                                 freq=freq,
                                 order=2,
                                 title=f"Scatterplot of {freq}-returns vs {benchmark}",
                                 ax=ax,
                                 **local_kwargs)

    def plot_performance_attribution(self,
                                     portfolio_ids: List[int] = (0, ),
                                     time_period: da.TimePeriod = None,
                                     attribution_metric: AttributionMetric = AttributionMetric.PNL,
                                     ax: plt.Subplot = None,
                                     **kwargs
                                     ) -> None:
        datas = []
        for portfolio_id in portfolio_ids:
            datas.append(self.portfolio_datas[portfolio_id].get_performance_data(attribution_metric=attribution_metric,
                                                                                 time_period=time_period))
        data = pd.concat(datas, axis=1)
        data = data.sort_values(data.columns[0], ascending=False)
        kwargs = sop.update_kwargs(kwargs=kwargs,
                                         new_kwargs={'ncol': len(data.columns),
                                                     'legend_loc': 'upper center',
                                                     'bbox_to_anchor': (0.5, 1.05),
                                                     'x_rotation': 90})
        data = data.replace({0.0: np.nan}).dropna()
        plot_bars(df=data,
                  skip_y_axis=True,
                  title=f"{attribution_metric.title}",
                  stacked=False,
                  ax=ax,
                  **kwargs)

    def plot_performance_periodic_table(self,
                                        portfolio_id: int = 0,
                                        time_period: da.TimePeriod = None,
                                        attribution_metric: AttributionMetric = AttributionMetric.INST_PNL,
                                        freq: str = 'M',
                                        ax: plt.Subplot = None,
                                        **kwargs
                                        ) -> None:

        inst_returns = self.portfolio_datas[portfolio_id].get_attribution_table_by_instrument(time_period=time_period)

        total_return = inst_returns.sum(axis=1).rename('Sum Total')

        strategy_nav = self.portfolio_datas[portfolio_id].get_portfolio_nav(time_period=time_period)
        nav_return = ret.to_returns(prices=strategy_nav, freq=freq, drop_first=True).rename('Nav Total')
        returns = pd.concat([inst_returns, total_return, nav_return], axis=1)
        total_returns = ((1.0+returns).cumprod(axis=0).iloc[-1, :]-1.0).rename('Period Total')
        returns = returns.append(total_returns)

        rhe.plot_periodic_returns_table(returns=returns,
                                        title=f"{strategy_nav.name} Attribution by Instrument",
                                        freq=freq,
                                        ax=ax,
                                        **kwargs)

    def plot_composite_table(self,
                             portfolio_id: int = 0,
                             date_format: str = '%b-%y',
                             time_period: da.TimePeriod = None,
                             perf_params: PerfParams = PERF_PARAMS,
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> plt.Figure:
        """
        plot performqances with weights
        """
        perf_columns: List[PerfStat] = rpt.BENCHMARK_TABLE_COLUMNS
        strategy_nav = self.portfolio_datas[portfolio_id].get_portfolio_nav(time_period=time_period)
        prices = pd.concat([strategy_nav, self.portfolio_datas[portfolio_id].get_instruments_navs(time_period=time_period)], axis=1)
        ra_perf_table = rpt.compute_ra_perf_table_with_benchmark(prices=prices,
                                                                 benchmark=str(strategy_nav.name),
                                                                 perf_params=perf_params)
        data = pd.DataFrame(data=prices.columns, index=ra_perf_table.index, columns=['Assets'])
        for perf_column in perf_columns:
            data[perf_column.to_str()] = dfs.series_to_str(ds=ra_perf_table[perf_column.to_str()],
                                                           var_format=perf_column.to_format(**kwargs))

        df = self.portfolio_datas[portfolio_id].weights
        if time_period is not None:
            df = time_period.locate(df)

        exposures = dfs.timeseries_df_to_str(df=df,
                                             freq='Q',
                                             var_format='{:.2%}',
                                             date_format=date_format,
                                             transpose=True)
        joint = pd.concat([data, exposures], axis=1)

        fig = ptb.plot_df_table(df=joint,
                                add_index_as_column=False,
                                transpose=False,
                                ax=ax,
                                **kwargs)
        return fig

    def plot_regime_data(self,
                         benchmark: str,
                         is_grouped: bool = False,
                         portfolio_idx: int = 0,
                         regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                         time_period: da.TimePeriod = None,
                         var_format: Optional[str] = None,
                         is_conditional_sharpe: bool = True,
                         perf_params: PerfParams = PERF_PARAMS,
                         regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                         legend_loc: Optional[str] = 'upper center',
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> None:
        if is_grouped:
            prices = self.get_ac_navs(portfolio_idx=portfolio_idx, benchmark=benchmark, time_period=time_period)
            title = f"Sharpe ratio decomposition by Asset Group to {benchmark} Bear/Normal/Bull regimes"
        else:
            prices = self.get_navs(benchmark=benchmark, time_period=time_period)
            title = f"Sharpe ratio decomposition by Strategies to {benchmark} Bear/Normal/Bull regimes"
        if var_format is None:
            if regime_data_to_plot == RegimeData.REGIME_SHARPE:
                var_format = '{:.2f}'
            else:
                var_format = '{:.2%}'
        regime_classifier = rcl.BenchmarkReturnsQuantilesRegime(regime_params=regime_params)
        qis.plots.derived.regime_data.plot_regime_data(regime_classifier=regime_classifier,
                                                       prices=prices,
                                                       benchmark=benchmark,
                                                       is_conditional_sharpe=is_conditional_sharpe,
                                                       regime_data_to_plot=regime_data_to_plot,
                                                       var_format=var_format,
                                                       regime_params=regime_params,
                                                       legend_loc=legend_loc,
                                                       perf_params=perf_params,
                                                       title=title,
                                                       ax=ax,
                                                       **kwargs)



"""
compute risk-adjusted performance tables
"""
# packages
import numpy as np
import pandas as pd
from scipy.stats import kurtosis, skew
from typing import Callable, Union, Dict, Tuple, Any, Optional
from collections import OrderedDict
from enum import Enum

# qis
import qis.utils.dates as da
import qis.utils.ols as ols
import qis.perfstats.returns as ret
from qis.perfstats.config import PerfStat, PerfParams

STANDARD_TABLE_COLUMNS = (PerfStat.START_DATE,
                          PerfStat.END_DATE,
                          PerfStat.PA_RETURN,
                          PerfStat.VOL,
                          PerfStat.SHARPE,
                          PerfStat.MAX_DD,
                          PerfStat.MAX_DD_VOL,
                          PerfStat.SKEWNESS,
                          PerfStat.KURTOSIS)

LN_TABLE_COLUMNS = (PerfStat.START_DATE,
                    PerfStat.END_DATE,
                    PerfStat.TOTAL_RETURN,
                    PerfStat.PA_RETURN,
                    PerfStat.AN_LOG_RETURN,
                    PerfStat.VOL,
                    PerfStat.SHARPE,
                    PerfStat.SHARPE_LOG_AN,
                    PerfStat.MAX_DD,
                    PerfStat.MAX_DD_VOL,
                    PerfStat.SKEWNESS,
                    PerfStat.KURTOSIS)

LN_BENCHMARK_TABLE_COLUMNS = (PerfStat.START_DATE,
                              PerfStat.END_DATE,
                              PerfStat.TOTAL_RETURN,
                              PerfStat.PA_RETURN,
                              PerfStat.AN_LOG_RETURN,
                              PerfStat.VOL,
                              PerfStat.SHARPE,
                              PerfStat.SHARPE_LOG_AN,
                              PerfStat.MAX_DD,
                              PerfStat.MAX_DD_VOL,
                              PerfStat.SKEWNESS,
                              PerfStat.KURTOSIS,
                              PerfStat.ALPHA,
                              PerfStat.BETA,
                              PerfStat.R2)

LN_BENCHMARK_TABLE_COLUMNS_SHORT = (PerfStat.TOTAL_RETURN,
                                    PerfStat.PA_RETURN,
                                    PerfStat.AN_LOG_RETURN,
                                    PerfStat.VOL,
                                    PerfStat.SHARPE,
                                    PerfStat.SHARPE_LOG_AN,
                                    PerfStat.MAX_DD,
                                    PerfStat.MAX_DD_VOL,
                                    PerfStat.SKEWNESS,
                                    PerfStat.ALPHA,
                                    PerfStat.BETA,
                                    PerfStat.R2)

EXTENDED_TABLE_COLUMNS = (PerfStat.START_DATE,
                          PerfStat.END_DATE,
                          PerfStat.START_PRICE,
                          PerfStat.END_PRICE,
                          PerfStat.TOTAL_RETURN,
                          PerfStat.PA_RETURN,
                          PerfStat.VOL,
                          PerfStat.SHARPE,
                          PerfStat.SHARPE_EXCESS,
                          PerfStat.MAX_DD,
                          PerfStat.MAX_DD_VOL,
                          PerfStat.SKEWNESS,
                          PerfStat.KURTOSIS)

COMPACT_TABLE_COLUMNS = (PerfStat.TOTAL_RETURN,
                         PerfStat.PA_RETURN,
                         PerfStat.VOL,
                         PerfStat.SHARPE,
                         PerfStat.MAX_DD,
                         PerfStat.MAX_DD_VOL,
                         PerfStat.SKEWNESS)

SMALL_TABLE_COLUMNS = (PerfStat.TOTAL_RETURN,
                       PerfStat.PA_RETURN,
                       PerfStat.VOL,
                       PerfStat.SHARPE,
                       PerfStat.MAX_DD)

BENCHMARK_TABLE_COLUMNS = (PerfStat.PA_RETURN,
                           PerfStat.VOL,
                           PerfStat.SHARPE,
                           PerfStat.MAX_DD,
                           PerfStat.MAX_DD_VOL,
                           PerfStat.SKEWNESS,
                           PerfStat.ALPHA_AN,
                           PerfStat.BETA,
                           PerfStat.R2)

BENCHMARK_TABLE_COLUMNS2 = (PerfStat.TOTAL_RETURN,
                            PerfStat.PA_RETURN,
                            PerfStat.VOL,
                            PerfStat.SHARPE,
                            PerfStat.MAX_DD,
                            PerfStat.MAX_DD_VOL,
                            PerfStat.SKEWNESS,
                            PerfStat.ALPHA_AN,
                            PerfStat.BETA,
                            PerfStat.R2)


def compute_performance_table(prices: Union[pd.DataFrame, pd.Series],
                              perf_params: PerfParams,
                              ) -> pd.DataFrame:
    """
    get performances for assets in prices
    """
    if not isinstance(prices, pd.DataFrame):
        raise TypeError(f"must be pd.Dataframe")

    if perf_params is None:
        perf_params = PerfParams()

    dict_data = OrderedDict()
    for asset in prices:
        asset_data = prices[asset].dropna()  # force drop na
        return_dict = ret.compute_returns_dict(prices=asset_data, perf_params=perf_params)
        dict_data[asset] = return_dict
        # keys will be rows = asset, column = keys in return_dict
    data = pd.DataFrame.from_dict(data=dict_data, orient='index')
    return data


def compute_risk_table(prices: pd.DataFrame,
                       perf_params: PerfParams = None
                       ) -> pd.DataFrame:
    """
    compute price data for returns statistics
    prices is limited to [sample_date_0,...,sample_date_N]
    """
    if not isinstance(prices, pd.DataFrame):
        raise TypeError(f"prices must be dataframe.")

    if perf_params is None:
        perf_params = PerfParams()

    sampled_prices = ret.prices_at_freq(prices=prices, freq=perf_params.freq_vol)

    # drawdowns are computed using its own freq
    if perf_params.freq_vol == perf_params.freq_drawdown:
        dd_sampled_prices = sampled_prices
    else:
        dd_sampled_prices = ret.prices_at_freq(prices=prices, freq=perf_params.freq_drawdown)

    vol_dt = np.sqrt(da.infer_an_from_data(data=sampled_prices))
    dict_data = OrderedDict()
    for asset in sampled_prices:
        sampled_price = sampled_prices[asset].dropna()
        if len(sampled_price.index) > 1:
            sampled_returns = ret.to_returns(prices=sampled_price, return_type=perf_params.return_type, drop_first=True)
            nd_sampled_returns = sampled_returns.to_numpy()
            vol = vol_dt * np.std(nd_sampled_returns, ddof=1)

            asset_dict = {PerfStat.VOL.to_str(): vol,
                          PerfStat.AVG_LOG_RETURN.to_str(): np.nanmean(sampled_returns),
                          PerfStat.START_DATE.to_str(): sampled_price.index[0],
                          PerfStat.END_DATE.to_str(): sampled_price.index[-1],
                          PerfStat.NUM_OBS.to_str(): len(nd_sampled_returns)
                          }
            # compute max dd on business day schedule
            max_dd = compute_max_dd(prices=dd_sampled_prices[asset].dropna())
            rel_returns = sampled_price.pct_change().dropna()
            asset_dict.update({
                PerfStat.MAX_DD.to_str(): max_dd,
                PerfStat.MAX_DD_VOL.to_str(): max_dd / vol if vol > 0.0 else 0.0,
                PerfStat.WORST.to_str(): np.min(rel_returns),
                PerfStat.BEST.to_str(): np.max(rel_returns),
                PerfStat.SKEWNESS.to_str(): skew(nd_sampled_returns, bias=False),
                PerfStat.KURTOSIS.to_str(): kurtosis(nd_sampled_returns, bias=False)
            })
        else:  # no required data
            asset_dict = {PerfStat.VOL.to_str(): np.nan,
                          PerfStat.AVG_LOG_RETURN.to_str(): np.nan,
                          PerfStat.START_DATE.to_str(): np.nan,
                          PerfStat.END_DATE.to_str(): np.nan,
                          PerfStat.NUM_OBS.to_str(): 0
                          }
        dict_data[asset] = asset_dict

        # keys will be rows = asset, column = keys in return_dict
    data = pd.DataFrame.from_dict(data=dict_data, orient='index')

    return data


def compute_ra_perf_table(prices: Union[pd.DataFrame, pd.Series],
                          perf_params: PerfParams = None
                          ) -> pd.DataFrame:
    if perf_params is None:
        perf_params = PerfParams(freq=pd.infer_freq(prices.index))

    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    perf_table = compute_performance_table(prices=prices, perf_params=perf_params)

    # is we only need sharpe we only comptute vol without higher order risk
    risk_table = compute_risk_table(prices=prices, perf_params=perf_params)

    # get vol and compute risk adjusted performance
    vol = risk_table[PerfStat.VOL.to_str()]
    perf_table[PerfStat.SHARPE.to_str()] = perf_table[PerfStat.PA_RETURN.to_str()] / vol
    perf_table[PerfStat.SHARPE_LOG_AN.to_str()] = perf_table[PerfStat.AN_LOG_RETURN.to_str()] / vol
    perf_table[PerfStat.SHARPE_AVG.to_str()] = perf_table[PerfStat.AVG_AN_RETURN.to_str()] / vol  # computed in risk
    perf_table[PerfStat.SHARPE_EXCESS.to_str()] = perf_table[PerfStat.PA_EXCESS_RETURN.to_str()] / vol
    perf_table[PerfStat.SHARPE_APR.to_str()] = perf_table[PerfStat.APR.to_str()] / vol

    # merge the two meta on dates
    ra_perf_table = pd.merge(left=perf_table, right=risk_table,
                             left_index=True, right_index=True, how='inner',
                             suffixes=(None, "_y"))
    return ra_perf_table


def compute_ra_perf_table_with_benchmark(prices: pd.DataFrame,
                                         benchmark: str,
                                         perf_params: PerfParams = None,
                                         is_log_returns: bool = False,
                                         **kwargs
                                         ) -> pd.DataFrame:
    assert benchmark in prices.columns
    if perf_params is None:
        perf_params = PerfParams(freq=pd.infer_freq(prices.index))

    ra_perf_table = compute_ra_perf_table(prices=prices, perf_params=perf_params)

    returns = ret.to_returns(prices=prices, freq=perf_params.freq_reg, is_log_returns=is_log_returns)
    alphas, betas, r2 = {}, {}, {}
    for column in returns.columns:
        joint_data = returns[[benchmark, column]].dropna()
        if joint_data.empty or len(joint_data.index) < 2:
            alphas[column], betas[column], r2[column] = np.nan, np.nan, np.nan
        else:
            alphas[column], betas[column], r2[column] = ols.estimate_ols_alpha_beta(x=joint_data.iloc[:, 0],
                                                                                    y=joint_data.iloc[:, 1])

            # get vol and compute risk adjusted performance
    ra_perf_table[PerfStat.ALPHA.to_str()] = pd.Series(alphas)
    ra_perf_table[PerfStat.ALPHA_AN.to_str()] = pd.Series(alphas)
    ra_perf_table[PerfStat.BETA.to_str()] = pd.Series(betas)
    ra_perf_table[PerfStat.R2.to_str()] = pd.Series(r2)

    return ra_perf_table


def compute_desc_freq_table(df: pd.DataFrame,
                            freq: str = 'A',
                            agg_func: Callable = np.sum
                            ) -> pd.DataFrame:
    """
    for time series pandas, aggregate by freq and produce descriprive data
    """
    freq_data = df.resample(freq).agg(agg_func)

    # drop na rows for all
    freq_data = freq_data.dropna(axis=0, how='any')

    data_values = freq_data.to_numpy()
    data_table = pd.DataFrame(index=freq_data.columns)
    data_table[PerfStat.AVG.to_str()] = np.nanmean(data_values, axis=0)
    data_table[PerfStat.STD.to_str()] = np.nanstd(data_values, ddof=1, axis=0)
    data_table[PerfStat.QUANT_M_1STD.to_str()] = np.nanquantile(data_values, q=0.16, axis=0)
    data_table[PerfStat.MEDIAN.to_str()] = np.mean(data_values, axis=0)
    data_table[PerfStat.QUANT_P1_STD.to_str()] = np.nanquantile(data_values, q=0.84, axis=0)

    return data_table


def compute_te_ir_errors(return_diffs: pd.DataFrame) -> Tuple[pd.Series, pd.Series]:
    """
    compute information ratio from return diffs
    """
    vol_dt = np.sqrt(da.infer_an_from_data(return_diffs))
    avg = np.nanmean(return_diffs, axis=0)
    vol = np.nanstd(return_diffs, axis=0, ddof=1)
    ir = vol_dt * np.divide(avg, vol, where=np.greater(vol, 0.0))
    te = pd.Series(vol_dt * vol, index=return_diffs.columns, name=PerfStat.TE.to_str())
    ir = pd.Series(ir, index=return_diffs.columns, name=PerfStat.IR.to_str())
    return te, ir


def compute_info_ratio_table(return_diffs_dict: Dict[str, pd.DataFrame]) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    compute information ratio from return diffs
    """
    te_ac_datas = []
    ir_ac_datas = []
    for ac, data in return_diffs_dict.items():
        te, ir = compute_te_ir_errors(return_diffs=data)
        te_ac_datas.append(te.rename(ac))
        ir_ac_datas.append(ir.rename(ac))
    te_table = pd.concat(te_ac_datas, axis=1)
    ir_table = pd.concat(ir_ac_datas, axis=1)
    return te_table, ir_table


def compute_drawdown(price: pd.Series) -> pd.Series:
    if not isinstance(price, pd.Series):
        print(price)
        raise TypeError(f"in compute_max_dd: path_data must be series")

    max_dd = np.zeros_like(price)
    last_peak = -np.inf

    for idx, price_ in enumerate(price):
        if not np.isnan(price_):
            if price_ > last_peak:
                last_peak = price_
            if not np.isclose(last_peak, 0.0):
                max_dd[idx] = price_ / last_peak - 1.0
        else:
            max_dd[idx] = np.nan

    max_dd = pd.Series(data=max_dd, index=price.index, name=price.name)

    return max_dd


def compute_time_under_water(price: pd.Series) -> Tuple[pd.Series, pd.Series]:
    if not isinstance(price, pd.Series):
        raise TypeError(f"in compute_time_under_water: path_data must be series")

    max_dd = np.zeros_like(price)
    dd_times = np.zeros_like(price.index)

    last_peak = -np.inf
    last_dd, last_dd_time = 0.0, price.index[0]
    dd_times[0] = last_dd_time
    for idx, (index, price_) in enumerate(price.items()):
        if not np.isnan(price_):
            if price_ > last_peak:
                last_peak = price_
                last_dd_time = index
            last_dd = price_ / last_peak - 1.0
            max_dd[idx] = last_dd
            dd_times[idx] = last_dd_time
        else:
            max_dd[idx] = last_dd
            dd_times[idx] = last_dd_time

    max_dd = pd.Series(data=max_dd, index=price.index, name=price.name)
    dd_times = (price.index - pd.DatetimeIndex(dd_times)).days
    time_under_water = pd.Series(data=dd_times, index=price.index, name=price.name)

    return max_dd, time_under_water


def compute_drawdown_data(prices: Union[pd.DataFrame, pd.Series]) -> Union[pd.DataFrame, pd.Series]:
    if isinstance(prices, pd.Series):
        drawdown = compute_drawdown(price=prices)
    elif isinstance(prices, pd.DataFrame):
        if len(prices.columns) > 1 and prices.columns.duplicated().any():
            raise ValueError(f"dublicated columns = {prices[prices.columns.duplicated()]}")
        drawdowns = []
        for asset_ in prices:
            drawdowns.append(compute_drawdown(price=prices[asset_]))
        drawdown = pd.concat(drawdowns, axis=1)
    else:
        raise ValueError(f"unsuported type {type(prices)}")
    return drawdown


def compute_drawdown_time_under_water(prices: Union[pd.DataFrame, pd.Series]
                                      ) -> Tuple[Union[pd.DataFrame, pd.Series], Union[pd.DataFrame, pd.Series]]:
    """
    compute joint data of drawdown and time under water
    """
    if isinstance(prices, pd.Series):
        drawdown, time_under_water = compute_time_under_water(price=prices)
    else:
        drawdowns = []
        time_under_waters = []
        for asset in prices.columns:
            drawdown, time_under_water = compute_time_under_water(price=prices[asset])
            drawdowns.append(drawdown)
            time_under_waters.append(time_under_water)
        drawdown = pd.concat(drawdowns, axis=1)
        time_under_water = pd.concat(time_under_waters, axis=1)

    return drawdown, time_under_water


def compute_max_dd(prices: Union[pd.DataFrame, pd.Series]) -> np.ndarray:
    """
    compute realized max drawdown
    """
    max_dd_data = compute_drawdown_data(prices=prices)
    max_dds = np.min(max_dd_data.to_numpy(), axis=0)
    return max_dds


def compute_avg_max_dd(ds: pd.Series,
                       is_max: bool = True,
                       q: float = 0.1
                       ) -> (float, float, float, float):
    """
    compute dd statistics
    """
    if is_max:
        nan_data = np.where(ds.to_numpy() >= 0, ds.to_numpy(), np.nan)
    else:
        nan_data = np.where(ds.to_numpy() <= 0, ds.to_numpy(), np.nan)

    avg = np.nanmean(nan_data)
    if is_max:
        quant = np.nanquantile(nan_data, 1.0 - q)
        nmax = np.nanmax(nan_data)
    else:
        quant = np.nanquantile(nan_data, q)
        nmax = np.nanmin(nan_data)

    last = ds.iloc[-1]

    return avg, quant, nmax, last


def compute_drawdowns_stats_table(price: pd.Series,
                                  max_num: Optional[int] = None,
                                  freq: Optional[str] = 'D'  # need to rebase to calendar days
                                  ) -> pd.DataFrame:
    """
    compute drawdown statistics table
    split drawdown time series into block
    rank block by max dradown and compute stats for the n-top blocks
    """
    if freq is not None:
        price = price.asfreq(freq, method='ffill')
    max_dd, time_under_water = compute_time_under_water(price=price)
    max_dd = max_dd.replace({0.0: np.nan})
    time_under_water = time_under_water.replace({0.0: np.nan})

    # Convert to sparse then query index to find block locations
    joint = pd.concat([max_dd.rename('max_dd'), time_under_water.rename('days'), price], axis=1)

    def process_bslice(bslice: pd.DataFrame) -> Dict[str, Any]:
        max_idx = np.argmin(bslice['max_dd'].to_numpy())
        out = dict(start=bslice.index[0],
                   trough=bslice.index[max_idx],
                   end=bslice.index[-1],
                   max_dd=bslice['max_dd'].iloc[max_idx],
                   days_dd=bslice['days'].iloc[-1],
                   days_to_trough=bslice['days'].iloc[max_idx],
                   days_recovery=bslice['days'].iloc[-1]-bslice['days'].iloc[max_idx],
                   peak=bslice[price.name].iloc[0],
                   bottom=bslice[price.name].iloc[max_idx],
                   recovery=bslice[price.name].iloc[-1])
        return out

    # split max_dd to series according to nan blocs
    sparse_ts = max_dd.astype(pd.SparseDtype('float'))
    # we need to use .values.sp_index.to_block_index() in this version of pandas
    outputs = {}
    for bstart, blength in zip(sparse_ts.values.sp_index.to_block_index().blocs,
                              sparse_ts.values.sp_index.to_block_index().blengths):
        bslice = joint.iloc[bstart: (bstart + blength - 1)]
        if not bslice.empty:
            outputs[bstart] = process_bslice(bslice=bslice)

    df = pd.DataFrame.from_dict(outputs, orient='index')
    df = df.sort_values(by='max_dd')
    if max_num is not None:
        df = df.iloc[:max_num, :]
    return df


class UnitTests(Enum):
    RA_PERF_TABLE = 1
    DRAWDOWN = 2
    DRAWDOWN_STATS_TABLE = 3


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.RA_PERF_TABLE:
        perf_params = PerfParams(freq='B')
        table = compute_ra_perf_table(prices=prices, perf_params=perf_params)
        print(table)
        print(table.columns)

    elif unit_test == UnitTests.DRAWDOWN:
        dd_data = compute_drawdown(price=prices['SPY'])
        print(dd_data)

    elif unit_test == UnitTests.DRAWDOWN_STATS_TABLE:
        df = compute_drawdowns_stats_table(price=prices['SPY'])
        print(df)


if __name__ == '__main__':

    unit_test = UnitTests.DRAWDOWN_STATS_TABLE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




# packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from typing import Tuple, List, Dict
from enum import Enum

# qis
import qis.utils.dates as da
import qis.perfstats.returns as ret
import qis.plots.time_series as pts

# models
import qis.models.linear.ewm as ewm
import qis.models.stats.gaussian_mixture as gm
from qis.models.linear.corr_cov_matrix import compute_masked_covar_corr, matrix_regularization

# portfolio
import qis.portfolio.optimization.opt_solvers as ops
import qis.portfolio.optimization.qp_solvers as qup
from qis.portfolio.optimization.qp_solvers import PortfolioObjective, max_portfolio_sharpe_qp
import qis.portfolio.backtester as bp
from qis.portfolio.portfolio_data import PortfolioData


def solve_optimal_weights_ewm_covar(prices: pd.DataFrame,
                                     portfolio_objective: PortfolioObjective = PortfolioObjective.MIN_VAR,
                                     fixed_weights: Dict[str, float] = None,  # fixed for principal portfolios
                                     weight_mins: np.ndarray = None,
                                     weight_maxs: np.ndarray = None,
                                     target_vol: float = None,
                                     rebalancing_freq: str = 'Q',
                                     ewm_lambda: float = 0.97,
                                     is_regularize: bool = False,
                                     returns_freq: str = None,
                                     is_log_returns: bool = True,
                                     budget: np.ndarray = None
                                     ) -> pd.DataFrame:

    returns = ret.to_returns(prices=prices,
                             is_log_returns=is_log_returns,
                             freq=returns_freq,
                             ffill_nans=True,
                             include_end_date=False)

    # drift adjusted returns
    returns_np = returns.to_numpy()
    x = returns_np # - ewm.ewm_nd(a=returns_np, ewm_lambda=ewm_lambda)
    covar_tensor_txy = ewm.compute_ewm_covar_tensor(a=x, ewm_lambda=ewm_lambda)

    rebalancing_schedule = da.generate_rebalancing_indicators(df=returns, freq=rebalancing_freq)
    columns = prices.columns.to_list()
    n = len(columns)

    if fixed_weights is not None:
        if weight_mins is None:
            weight_mins = np.zeros(n)
        if weight_maxs is None:
            weight_maxs = np.ones(n)
        for asset, weight in fixed_weights.items():
            idx = columns.index(asset)
            weight_mins[idx], weight_maxs[idx] = weight, weight

    an_factor = da.infer_an_from_data(data=returns)
    weights = {}
    for idx, (date, value) in enumerate(rebalancing_schedule.items()):
        if value:
            covar = an_factor*covar_tensor_txy[idx]
            if is_regularize:
                covar = matrix_regularization(covar=covar)

            if portfolio_objective == PortfolioObjective.EQUAL_RISK_CONTRIBUTION:
                if target_vol is None:
                    weights[date] = ops.solve_equal_risk_contribution(covar=covar,
                                                                     budget=budget,
                                                                     weight_mins=weight_mins,
                                                                     weight_maxs=weight_maxs)
                else:
                    weights[date] = ops.solve_risk_parity_constr_vol(covar=covar,
                                                                    target_vol=target_vol)

            elif portfolio_objective == PortfolioObjective.MAX_DIVERSIFICATION:
                weights[date] = ops.solve_max_diversification(covar=covar,
                                                             weight_mins=weight_mins,
                                                             weight_maxs=weight_maxs)

            elif portfolio_objective == PortfolioObjective.RISK_PARITY_ALT:
                weights[date] = ops.solve_risk_parity_alt(covar=covar)

            else:
                weights[date] = qup.maximize_portfolio_objective_qp(portfolio_objective=portfolio_objective,
                                                                    covar=covar,
                                                                    is_gross_notional_one=True,
                                                                    is_long_only=True,
                                                                    weight_mins=weight_mins,
                                                                    weight_maxs=weight_maxs)
    weights = pd.DataFrame.from_dict(weights, orient='index', columns=returns.columns)
    return weights


def run_rolling_mixure_portfolios(prices: pd.DataFrame,
                                  recalib_freq: str = 'A',
                                  roll_window: int = 5,
                                  returns_freq: str = 'M',
                                  is_log_returns: bool = True,
                                  carra: float = 0.5,
                                  n_components: int = 3,
                                  weight_mins: np.ndarray = None,
                                  weight_maxs: np.ndarray = None,
                                  ticker: str = None
                                  ) -> PortfolioData:

    rets = ret.to_returns(prices=prices, is_log_returns=is_log_returns, drop_first=True, freq=returns_freq)

    dates_schedule = da.generate_dates_schedule(time_period=da.get_time_period(df=rets),
                                                freq=recalib_freq,
                                                include_start_date=True,
                                                include_end_date=False)
    _, scaler = da.get_period_days(freq=returns_freq)
    weights = {}
    for idx, end in enumerate(dates_schedule[1:]):
        if idx >= roll_window-1:
            period = da.TimePeriod(dates_schedule[idx - roll_window+1], end)
            # period.print()
            rets_ = period.locate(rets).to_numpy()
            params = gm.fit_gaussian_mixture(x=rets_, n_components=n_components, scaler=scaler)
            # print(params)
            weights[end] = ops.solve_cara_mixture(means=params.means,
                                                  covars=params.covars,
                                                  probs=params.probs,
                                                  carra=carra,
                                                  weight_mins=weight_mins,
                                                  weight_maxs=weight_maxs)

    weights = pd.DataFrame.from_dict(weights, orient='index', columns=prices.columns)
    portfolio_out = bp.backtest_model_portfolio(prices=prices,
                                                weights=weights,
                                                is_rebalanced_at_first_date=True,
                                                ticker=ticker,
                                                is_output_portfolio_data=True)
    return portfolio_out


def run_rolling_erc_portfolios(prices: pd.DataFrame,
                               weight_mins: np.ndarray = None,
                               weight_maxs: np.ndarray = None,
                               time_period: da.TimePeriod = None,
                               portfolio_objective: PortfolioObjective = PortfolioObjective.EQUAL_RISK_CONTRIBUTION,
                               recalib_freq: str = 'Q',
                               ewm_lambda: float = 0.926,  #0.926,
                               returns_freq: str = 'W-WED',
                               budget: np.ndarray = None,
                               ticker: str = None
                               ) -> PortfolioData:

    weights = solve_optimal_weights_ewm_covar(prices=prices,
                                              portfolio_objective=portfolio_objective,
                                              weight_mins=weight_mins,
                                              weight_maxs=weight_maxs,
                                              rebalancing_freq=recalib_freq,
                                              returns_freq=returns_freq,
                                              ewm_lambda=ewm_lambda,
                                              budget=budget)
    if time_period is not None:
        prices = time_period.locate(prices)
        weights = time_period.locate(weights)
    portfolio_out = bp.backtest_model_portfolio(prices=prices.loc[weights.index[0]:, :],
                                                weights=weights,
                                                is_rebalanced_at_first_date=True,
                                                ticker=ticker,
                                                is_output_portfolio_data=True)
    return portfolio_out


def estimate_rolling_means_covar(prices: pd.DataFrame,
                                 returns_freq: str = 'M',
                                 recalib_freq: str = 'A',
                                 roll_window: int = 5,
                                 is_log_returns: bool = True,
                                 is_annualize: bool = True,
                                 is_regularize: bool = True,
                                 is_ewm_covar: bool = False
                                 ) -> Tuple[pd.DataFrame, List[np.ndarray]]:
    rets = ret.to_returns(prices=prices, is_log_returns=is_log_returns, drop_first=True, freq=returns_freq)

    dates_schedule = da.generate_dates_schedule(time_period=da.get_time_period(df=rets),
                                                freq=recalib_freq,
                                                include_start_date=True,
                                                include_end_date=False)

    if is_annualize:
        _, scaler = da.get_period_days(freq=returns_freq)
    else:
        scaler = 1.0
    means = {}
    covars = []
    covar0 = np.zeros((len(prices.columns), len(prices.columns)))
    for idx, end in enumerate(dates_schedule[1:]):
        if idx >= roll_window-1:
            period = da.TimePeriod(dates_schedule[idx - roll_window+1], end)
            # period.print()
            rets_ = period.locate(rets).to_numpy()
            means[end] = scaler*np.nanmean(rets_, axis=0)
            if is_ewm_covar:
                covar = ewm.compute_ewm_covar(a=rets_,
                                                ewm_lambda=0.846,
                                                covar0=covar0)
                covar0 = covar
            else:
                covar = compute_masked_covar_corr(returns=rets_, bias=True)

            if is_regularize:
                covar = matrix_regularization(covar=covar, cut=1e-5)

            covars.append(scaler * covar)
    means = pd.DataFrame.from_dict(means, orient="index")

    return means, covars


def run_rolling_mv_portfolios(prices: pd.DataFrame,
                              returns_freq: str = 'M',
                              recalib_freq: str = 'A',
                              roll_window: int = 5,
                              carra: float = 0.5,
                              time_period: da.TimePeriod = None,
                              is_log_returns: bool = True,
                              weight_mins: np.ndarray = None,
                              weight_maxs: np.ndarray = None,
                              ticker: str = None,
                              is_print_log: bool = True
                              ) -> PortfolioData:

    means, covars = estimate_rolling_means_covar(prices=prices,
                                                 returns_freq=returns_freq,
                                                 recalib_freq=recalib_freq,
                                                 roll_window=roll_window,
                                                 is_annualize=True,
                                                 is_log_returns=is_log_returns)
    weights = {}
    for index, covar in zip(means.index, covars):
        if np.isclose(carra, 0.0):
            weights[index] = max_portfolio_sharpe_qp(means=means.loc[index, :].to_numpy(),
                                                     covar=covar,
                                                     weight_mins=weight_mins,
                                                     weight_maxs=weight_maxs,
                                                     is_gross_notional_one=True,
                                                     is_print_log=is_print_log)
        else:
            weights[index] = ops.solve_cara(means=means.loc[index, :].to_numpy(), covar=covar,
                                            weight_mins=weight_mins,
                                            weight_maxs=weight_maxs,
                                            carra=carra)

    weights = pd.DataFrame.from_dict(weights, orient='index', columns=prices.columns)
    if time_period is not None:
        prices = time_period.locate(prices)
        weights = time_period.locate(weights)
    portfolio_out = bp.backtest_model_portfolio(prices=prices,
                                                weights=weights,
                                                is_rebalanced_at_first_date=True,
                                                ticker=ticker,
                                                is_output_portfolio_data=True)
    return portfolio_out


def estimate_rolling_mixture1(prices: pd.DataFrame,
                              returns_freq: str = 'M',
                              recalib_freq: str = 'A',
                              roll_window: int = 6,
                              n_components: int = 2,
                              is_log_returns: bool = True,
                              is_annualize: bool = True
                              ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:

    if len(prices.columns) > 1:
        raise ValueError
    rets = ret.to_returns(prices=prices, is_log_returns=is_log_returns, drop_first=True, freq=returns_freq)

    dates_schedule = da.generate_dates_schedule(time_period=da.get_time_period(df=rets),
                                                freq=recalib_freq,
                                                include_start_date=True,
                                                include_end_date=True)
    if is_annualize:
        _, scaler = da.get_period_days(freq=returns_freq)
    else:
        scaler = 1.0

    means, sigmas, probs = [], [], []
    for idx, end in enumerate(dates_schedule[1:]):
        if idx >= roll_window-1:
            period = da.TimePeriod(dates_schedule[idx - roll_window+1], end)
            # period.print()
            rets_ = period.locate(rets).to_numpy()
            params = gm.fit_gaussian_mixture(x=rets_, n_components=n_components, scaler=scaler)
            mean = np.stack(params.means, axis=0).T[0]
            std = np.sqrt(np.array([params.covars[0][0], params.covars[1][0]]))
            prob = params.probs
            ranks = mean.argsort().argsort()
            means.append(pd.DataFrame(mean[ranks].reshape(1, -1), index=[end]))
            sigmas.append(pd.DataFrame(std[ranks].reshape(1, -1), index=[end]))
            probs.append(pd.DataFrame(prob[ranks].reshape(1, -1), index=[end]))

    means = pd.concat(means)
    sigmas = pd.concat(sigmas)
    probs = pd.concat(probs)

    return means, sigmas, probs


class UnitTests(Enum):
    MIN_VAR = 1
    MIN_VAR_OVERLAY = 2
    RISK_PARITY = 3
    ROLLING_MEANS_COVAR = 4
    ROLLING_PORTFOLIOS = 5
    ROLLING_MIXTURES = 6
    MIXTURE_PORTFOLIOS = 7


def run_unit_test(unit_test: UnitTests):

    # data
    from qis.test_data import load_etf_data
    prices = load_etf_data()
    import qis.plots.stackplot as pst

    kwargs = dict(add_mean_levels=True,
                  is_yaxis_limit_01=True,
                  baseline='zero',
                  bbox_to_anchor=(0.4, 1.1),
                  legend_stats=pst.LegendStats.AVG_STD_LAST,
                  ncol=len(prices.columns)//3,
                  var_format='{:.0%}')

    if unit_test == UnitTests.MIN_VAR:
        weights = solve_optimal_weights_ewm_covar(prices=prices,
                                                  portfolio_objective=PortfolioObjective.MIN_VAR)
        pst.plot_stack(df=weights, **kwargs)

    elif unit_test == UnitTests.MIN_VAR_OVERLAY:
        fixed_weights = {'SPY': 0.6}
        weights = solve_optimal_weights_ewm_covar(prices=prices, fixed_weights=fixed_weights)
        pst.plot_stack(df=weights, **kwargs)

    elif unit_test == UnitTests.RISK_PARITY:
        weights = solve_optimal_weights_ewm_covar(prices=prices,
                                                  portfolio_objective=PortfolioObjective.EQUAL_RISK_CONTRIBUTION)
        pst.plot_stack(df=weights, **kwargs)

    elif unit_test == UnitTests.ROLLING_MEANS_COVAR:
        prices = prices[['SPY', 'TLT']].dropna()

        means, covars = estimate_rolling_means_covar(prices=prices, recalib_freq='A', roll_window=5)
        #  = estimate_rolling_data(prices=prices, recalib_freq='M', roll_window=60)

        vols = {}
        covs = {}
        for index, covar in zip(means.index, covars):
            vols[index] = pd.Series(np.sqrt(np.diag(covar)))
            covs[index] = pd.Series(np.extract(1 - np.eye(2), covar))
        vols = pd.DataFrame.from_dict(vols, orient='index')
        covs = pd.DataFrame.from_dict(covs, orient='index')
        print(vols)
        print(covs)

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(3, 1, figsize=(7, 12))
            pts.plot_time_series(df=means,
                                 var_format='{:.0%}',
                                 trend_line=pts.TrendLine.AVERAGE,
                                 legend_stats=pts.LegendStats.FIRST_AVG_LAST,
                                 ax=axs[0])
            pts.plot_time_series(df=vols,
                                 var_format='{:.0%}',
                                 trend_line=pts.TrendLine.AVERAGE,
                                 legend_stats=pts.LegendStats.FIRST_AVG_LAST,
                                 ax=axs[1])
            pts.plot_time_series(df=covs,
                                 var_format='{:.0%}',
                                 trend_line=pts.TrendLine.AVERAGE,
                                 legend_stats=pts.LegendStats.FIRST_AVG_LAST,
                                 ax=axs[2])

    elif unit_test == UnitTests.ROLLING_PORTFOLIOS:
        prices = prices.dropna()

        port_data = run_rolling_mv_portfolios(prices=prices, carra=0.0, recalib_freq='A', roll_window=5)
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(7, 12))
            pts.plot_time_series(df=port_data.weights,
                                 var_format='{:.0%}',
                                 legend_stats=pts.LegendStats.FIRST_AVG_LAST,
                                 ax=ax)

    elif unit_test == UnitTests.ROLLING_MIXTURES:
        prices = prices['SPY'].dropna()
        means, sigmas, probs = estimate_rolling_mixture1(prices=prices)
        print(means)

    elif unit_test == UnitTests.MIXTURE_PORTFOLIOS:
        prices = prices.dropna()
        port_data = run_rolling_mixure_portfolios(prices=prices, recalib_freq='Q', n_components=4, roll_window=4 * 5, carra=0.5)
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(7, 12))
            pts.plot_time_series(df=port_data.weights,
                                 var_format='{:.0%}',
                                 legend_stats=pts.LegendStats.FIRST_AVG_LAST,
                                 ax=ax)
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.RISK_PARITY

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# biult in
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import skew
from typing import Optional, Union, List, Tuple
from enum import Enum

# qis
import qis.utils.dates as da
import qis.utils.df_ops as dfo
import qis.utils.struct_ops as sop
import qis.perfstats.returns as ret
import qis.perfstats.perf_stats as pt
from qis.perfstats.config import PerfStat, PerfParams
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs

import qis.plots.derived.drawdowns as dra
import qis.plots.time_series as pts
import qis.plots.utils as put
from qis.plots.derived.regime_data import add_bnb_regime_shadows


class PerformanceLabel(Enum):
    NONE = 1
    SHARPE = 2
    DETAILED = 3
    WITH_SKEW = 4
    DETAILED_LOG = 5
    WITH_DD = 6
    WITH_DDVOL = 7
    PA_DETAILED = 8
    TOTAL = 9
    TOTAL_YTD = 10
    TOTAL_DETAILED = 11
    ARITHMETIC = 12
    TOTAL_ARITHMETIC = 13


def get_performance_labels(prices: Union[pd.DataFrame, pd.Series],
                           performance_label: PerformanceLabel = PerformanceLabel.DETAILED,
                           perf_params: PerfParams = None,
                           digits_to_show: int = 2,
                           sharpe_format: str = '{:.2f}',
                           **kwargs
                           ) -> List[str]:
    ra_perf_table = pt.compute_ra_perf_table(prices=prices, perf_params=perf_params)

    if digits_to_show == 2:
        ra_vol_vormat = '{:.2%}'
        vol_vormat = '{:.2f}'
    elif digits_to_show == 1:
        ra_vol_vormat = '{:.1%}'
        vol_vormat = '{:.1f}'
    else:
        ra_vol_vormat = '{:.0%}'
        vol_vormat = '{:.0f}'

    legend_labels = []
    for index in ra_perf_table.index:
        name = index if isinstance(index, str) else str(index)
        if performance_label == PerformanceLabel.NONE:
            label = f"{name}"
        elif performance_label == PerformanceLabel.SHARPE:
            label = f"{name}: Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}"
        elif performance_label == PerformanceLabel.DETAILED:
            label = (f"{name}: p.a.={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.PA_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}")
        elif performance_label == PerformanceLabel.WITH_SKEW:
            label = (f"{name}: p.a.={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.PA_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}, "
                     f"Skew={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SKEWNESS.to_str()])}")
        elif performance_label == PerformanceLabel.DETAILED_LOG:
            label = (f"{name}: an.={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.AN_LOG_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE_LOG_AN.to_str()])}")
        elif performance_label == PerformanceLabel.WITH_DD:
            label = (f"{name}: p.a.={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.PA_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}, "
                     f"MaxDD={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD.to_str()])}")
        elif performance_label == PerformanceLabel.WITH_DDVOL:
            label = (f"{name}: p.a.={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.PA_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}, "
                     f"MaxDD={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD.to_str()])}, "
                     f"MaxDD/vol={vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD_VOL.to_str()])}")
        elif performance_label == PerformanceLabel.PA_DETAILED:
            label = (f"{name}: P.a.={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.PA_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}, "
                     f"MaxDD={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD.to_str()])}, "
                     f"MaxDD/vol={vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD_VOL.to_str()])}, "
                     f"Skew={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SKEWNESS.to_str()])}")
        elif performance_label == PerformanceLabel.TOTAL:
            label = f"{name}: Total={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.TOTAL_RETURN.to_str()])}"
        elif performance_label == PerformanceLabel.TOTAL_YTD:
            label = (f"{name}: Total={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.TOTAL_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}, "
                     f"MaxDD={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD.to_str()])} ")
        elif performance_label == PerformanceLabel.TOTAL_DETAILED:
            label = (f"{name}: Total={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.TOTAL_RETURN.to_str()])}, "
                     f"p.a.={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.PA_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}, "
                     f"MaxDD={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD.to_str()])} ")
        elif performance_label == PerformanceLabel.ARITHMETIC:
            label = (f"{name}: Total={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.TOTAL_RETURN.to_str()])}, "
                     f"MaxDD={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD.to_str()])} ")
        elif performance_label == PerformanceLabel.TOTAL_ARITHMETIC:
            label = (f"{name}: Total={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.TOTAL_RETURN.to_str()])}, "
                     f"vol={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.VOL.to_str()])}, "
                     f"Sharpe={sharpe_format.format(ra_perf_table.loc[index, PerfStat.SHARPE.to_str()])}, "
                     f"MaxDD={ra_vol_vormat.format(ra_perf_table.loc[index, PerfStat.MAX_DD.to_str()])} ")
        else:
            raise NotImplementedError(f"{performance_label}")
        legend_labels.append(label)

    return legend_labels


def plot_prices(prices: Union[pd.DataFrame, pd.Series],
                perf_params: PerfParams = None,
                regime_benchmark_str: str = None,  # to add regimes
                pivot_prices: pd.Series = None,
                regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                var_format: str = '{:,.1f}',
                digits_to_show: int = 1,
                sharpe_format: str = '{:.2f}',
                x_date_freq: str = 'Q',
                trend_line: put.TrendLine = put.TrendLine.NONE,
                is_log: bool = False,
                resample_freq: str = None,
                start_to_one: bool = True,
                end_to_one: bool = False,
                performance_label: PerformanceLabel = PerformanceLabel.DETAILED,
                title: str = None,
                ax: plt.Subplot = None,
                **kwargs
                ) -> plt.Figure:

    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    legend_labels = get_performance_labels(prices=prices,
                                           performance_label=performance_label,
                                           perf_params=perf_params,
                                           digits_to_show=digits_to_show,
                                           sharpe_format=sharpe_format,
                                           **kwargs)

    if resample_freq is not None:
        prices = prices.asfreq(resample_freq, method='ffill')

    if end_to_one:
        scaler = 1.0 / dfo.get_last_nonnan_values(df=prices)
        prices = prices.multiply(scaler)
    elif start_to_one:
        prices = prices.divide(dfo.get_first_nonnan_values(df=prices))

    fig = pts.plot_time_series(df=prices,
                               trend_line=trend_line,
                               var_format=var_format,
                               title=title,
                               legend_labels=legend_labels,
                               is_log=is_log,
                               x_date_freq=x_date_freq,
                               ax=ax,
                               **kwargs)

    if regime_benchmark_str is not None and regime_params is not None:
        add_bnb_regime_shadows(ax=ax,
                               data_df=prices,
                               pivot_prices=pivot_prices,
                               benchmark=regime_benchmark_str,
                               regime_params=regime_params)
    return fig


def plot_prices_with_dd(prices: Union[pd.DataFrame, pd.Series],
                        perf_params: PerfParams = None,
                        regime_benchmark_str: str = None,  # to add regimes
                        pivot_prices: pd.Series = None,
                        regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                        var_format: str = '{:,.1f}',
                        digits_to_show: int = 1,
                        sharpe_format: str = '{:.2f}',
                        x_date_freq: str = 'Q',
                        performance_label: PerformanceLabel = PerformanceLabel.WITH_DD,
                        is_log: bool = False,
                        remove_xticklabels_ax1: bool = True,
                        title: str = 'Performance',
                        dd_title: str = 'Running Drawdown',
                        dd_legend_type: dra.DdLegendType = dra.DdLegendType.NONE,
                        axs: List[plt.Subplot] = None,
                        **kwargs
                        ) -> plt.Figure:

    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    if axs is None:
        fig, axs = plt.subplots(2, 1)
    else:
        fig = None

    plot_prices(prices=prices,
                perf_params=perf_params,
                var_format=var_format,
                digits_to_show=digits_to_show,
                sharpe_format=sharpe_format,
                performance_label=performance_label,
                x_date_freq=x_date_freq,
                title=title,
                is_log=is_log,
                ax=axs[0],
                **kwargs)

    dra.plot_drawdown(prices=prices,
                      perf_params=perf_params,
                      dd_legend_type=dd_legend_type,
                      x_date_freq=x_date_freq,
                      title=dd_title,
                      ax=axs[1],
                      **kwargs)

    if remove_xticklabels_ax1:
        axs[0].set_xticklabels('')

    if regime_benchmark_str is not None and regime_params is not None:
        for ax in axs:
            add_bnb_regime_shadows(ax=ax,
                                   data_df=prices,
                                   pivot_prices=pivot_prices,
                                   benchmark=regime_benchmark_str,
                                   regime_params=regime_params,
                                   perf_params=perf_params)
    return fig


def plot_prices_with_fundamentals(prices: Union[pd.DataFrame, pd.Series],
                                  volumes: Union[pd.DataFrame, pd.Series],
                                  mcap: Union[pd.DataFrame, pd.Series],
                                  perf_params: PerfParams = None,
                                  regime_benchmark_str: str = None,  # to add regimes
                                  pivot_prices: pd.Series = None,
                                  regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                                  trend_line: put.TrendLine = put.TrendLine.AVERAGE,
                                  var_format: str = '{:,.1f}',
                                  digits_to_show: int = 2,
                                  sharpe_format: str = '{:.2f}',
                                  performance_label: PerformanceLabel = PerformanceLabel.WITH_DD,
                                  is_log: bool = False,
                                  title: str = None,
                                  dd_title: str = 'Running Drawdown',
                                  dd_legend_type: dra.DdLegendType = dra.DdLegendType.NONE,
                                  axs: List[plt.Subplot] = None,
                                  **kwargs
                                  ) -> plt.Figure:

    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    if axs is None:
        fig, axs = plt.subplots(3, 1)
    else:
        fig = None

    plot_prices(prices=prices,
                perf_params=perf_params,
                var_format=var_format,
                digits_to_show=digits_to_show,
                sharpe_format=sharpe_format,
                performance_label=performance_label,
                is_log=is_log,
                ax=axs[0],
                **kwargs)

    dra.plot_drawdown(prices=prices,
                      perf_params=perf_params,
                      dd_legend_type=dd_legend_type,
                      title=dd_title,
                      ax=axs[1],
                      **kwargs)

    pts.plot_time_series_2ax(df1=volumes,
                             df2=mcap,
                             trend_line1=trend_line,
                             trend_line2=trend_line,
                             var_format=var_format,
                             title=title,
                             ax=axs[2],
                             **kwargs)

    axs[0].set_xticklabels('')
    axs[1].set_xticklabels('')

    if regime_benchmark_str is not None and regime_params is not None:
        for ax in axs:
            add_bnb_regime_shadows(ax=ax,
                                   data_df=prices,
                                   pivot_prices=pivot_prices,
                                   benchmark=regime_benchmark_str,
                                   regime_params=regime_params,
                                   perf_params=perf_params)
    return fig


def plot_prices_2ax(prices_ax1: Union[pd.DataFrame, pd.Series],
                    prices_ax2: Union[pd.DataFrame, pd.Series],
                    perf_params: PerfParams = None,
                    var_format: str = '{:,.1f}',
                    digits_to_show: int = 2,
                    sharpe_format: str = '{:.2f}',
                    trend_line: put.TrendLine = put.TrendLine.NONE,
                    is_logs: Tuple[bool, bool] = (False, False),
                    start_to_one: bool = True,
                    end_to_one: bool = False,
                    performance_label: PerformanceLabel = PerformanceLabel.DETAILED,
                    title: str = None,
                    ax: plt.Subplot = None,
                    **kwargs
                    ) -> plt.Figure:

    if isinstance(prices_ax1, pd.Series):
        prices_ax1 = prices_ax1.to_frame()
    if isinstance(prices_ax2, pd.Series):
        prices_ax2 = prices_ax2.to_frame()

    prices_ax1.columns = [f"{x} (left)" for x in prices_ax1.columns]
    prices_ax2.columns = [f"{x} (right)" for x in prices_ax2.columns]
    legend_labels1 = get_performance_labels(prices=prices_ax1,
                                            performance_label=performance_label,
                                            perf_params=perf_params,
                                            digits_to_show=digits_to_show,
                                            sharpe_format=sharpe_format,
                                            **kwargs)
    legend_labels2 = get_performance_labels(prices=prices_ax2,
                                            performance_label=performance_label,
                                            perf_params=perf_params,
                                            digits_to_show=digits_to_show,
                                            sharpe_format=sharpe_format,
                                            **kwargs)

    legend_labels = sop.to_flat_list(legend_labels1 + legend_labels2)

    if start_to_one:
        prices_ax1 = prices_ax1.divide(dfo.get_first_nonnan_values(df=prices_ax1))

    fig = pts.plot_time_series_2ax(df1=prices_ax1,
                                   df2=prices_ax2,
                                   trend_line=trend_line,
                                   var_format=var_format,
                                   title=title,
                                   legend_labels=legend_labels,
                                   is_logs=is_logs,
                                   ax=ax,
                                   **kwargs)
    return fig


def plot_rolling_sharpe(prices: pd.DataFrame,
                        is_sharpe: bool = True,
                        time_period: da.TimePeriod = None,
                        roll_periods: int = 260,
                        freq: str = None,
                        legend_stats: pts.LegendStats = pts.LegendStats.AVG_LAST,
                        var_format: str = '{:.2f}',
                        title: Optional[str] = None,
                        regime_benchmark_str: str = None,
                        pivot_prices: pd.Series = None,
                        regime_params: BenchmarkReturnsQuantileRegimeSpecs = BenchmarkReturnsQuantileRegimeSpecs(),
                        perf_params: PerfParams = None,
                        ax: plt.Subplot = None,
                        **kwargs
                        ) -> plt.Figure:

    if is_sharpe:
        df = compute_rolling_sharpes(prices=prices,
                                      roll_periods=roll_periods,
                                      freq=freq)
    else:
        df = compute_rolling_skew(prices=prices,
                                  roll_periods=roll_periods,
                                  freq=freq)

    if time_period is not None:
        df = time_period.locate(df)

    fig = pts.plot_time_series(df=df,
                               legend_stats=legend_stats,
                               var_format=var_format,
                               title=title,
                               ax=ax,
                               **kwargs)

    if regime_benchmark_str is not None and regime_params is not None:
        add_bnb_regime_shadows(ax=ax,
                               data_df=prices.reindex(index=df.index, method='ffill'),
                               pivot_prices=pivot_prices,
                               benchmark=regime_benchmark_str,
                               regime_params=regime_params,
                               perf_params=perf_params)

    return fig


def compute_rolling_sharpes(prices: Union[pd.Series, pd.DataFrame],
                            freq: Optional[str] = None,
                            roll_periods: int = 60,  # 5y * 12 month
                            ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=freq, is_log_returns=True, drop_first=False)
    saf = np.sqrt(da.infer_an_from_data(data=log_returns))
    sharpes = log_returns.rolling(roll_periods).apply(lambda x: compute_sharpe(x, saf=saf))
    return sharpes


def compute_sharpe(log_returns: Union[pd.Series, pd.DataFrame], saf: float = None) -> np.ndarray:
    if saf is None:
        saf = np.sqrt(da.infer_an_from_data(data=log_returns))
    mean = np.expm1(np.nanmean(log_returns.to_numpy()))
    vol = np.nanstd(log_returns.to_numpy(), ddof=1)
    if np.greater(vol, 0.0):
        sharpe = saf * mean / vol
    else:
        sharpe = np.nan
    return sharpe


def compute_rolling_skew(prices: Union[pd.Series, pd.DataFrame],
                         freq: Optional[str] = None,
                         roll_periods: int = 60,  # 5y * 12 month
                         ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=freq, is_log_returns=True, drop_first=False)
    skw = log_returns.rolling(roll_periods).apply(lambda x: compute_skew(x))
    return skw


def compute_skew(log_returns: Union[pd.Series, pd.DataFrame]) -> np.ndarray:
    skw = skew(log_returns.to_numpy(), axis=0, nan_policy='omit')
    return skw


class UnitTests(Enum):
    PRICE = 1
    PRICE_WITH_DD = 2


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.PRICE:
        perf_params = PerfParams(freq='B')
        plot_prices(prices=prices, perf_params=perf_params)

    elif unit_test == UnitTests.PRICE_WITH_DD:
        perf_params = PerfParams(freq='M')
        plot_prices_with_dd(prices=prices,
                            regime_benchmark_str=prices.columns[0],
                            perf_params=perf_params)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PRICE_WITH_DD

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
compute descriptive data table
"""
# packages
import numpy as np
import pandas as pd
from typing import Union
from scipy.stats import skew, kurtosis, percentileofscore, normaltest
from enum import Enum

# qis
import qis.utils as qu
from qis.perfstats.config import PerfStat


class DescTableType(Enum):
    NONE = 0
    SHORT = 1
    AVG_WITH_POSITIVE_PROB = 2
    WITH_POSITIVE_PROB = 3
    WITH_KURTOSIS = 4
    WITH_SCORE = 5
    EXTENSIVE = 6
    SKEW_KURTOSIS = 7


def compute_desc_table(df: Union[pd.DataFrame, pd.Series],
                       desc_table_type: DescTableType = DescTableType.SHORT,
                       var_format: str = '{:.2f}',
                       annualize_vol: bool = False,
                       is_add_tstat: bool = False,
                       norm_variable_display_type: str = '{:.1f}',  # for t-stsat
                       **kwargs
                       ) -> pd.DataFrame:
    """
    data corresponds to matrix of returns with index = time and columns = tickers
    data can contain nan in columns
    output is index = tickers, columns = descriptive data
    """
    if isinstance(df, pd.DataFrame):
        descriptive_table = pd.DataFrame(index=df.columns)
    elif isinstance(df, pd.Series):
        descriptive_table = pd.DataFrame(index=df.name)
        df = df.to_frame()
    else:
        raise TypeError(f"unsupported data type = {type(df)}")

    data_np = df.to_numpy()
    mean = np.nanmean(data_np, axis=0)
    std = np.nanstd(data_np, ddof=1, axis=0)

    descriptive_table[PerfStat.AVG.to_str()] = [var_format.format(x) for x in mean]

    if annualize_vol:
        an_factor = qu.infer_an_from_data(data=df)
        vol = std * np.sqrt(an_factor)
        descriptive_table[PerfStat.STD_AN.to_str()] = [var_format.format(x) for x in vol]
    else:
        an_factor = 1.0
        vol = std
        descriptive_table[PerfStat.STD.to_str()] = [var_format.format(x) for x in std]

    if is_add_tstat:
        an_mean = an_factor * mean
        tstats = np.divide(an_mean, vol, where=np.greater(an_mean, 0.0))
        descriptive_table[PerfStat.T_STAT.to_str()] = [norm_variable_display_type.format(x) for x in tstats]

    nan_policy = 'omit'  # skip nans
    if desc_table_type == desc_table_type.SHORT:
        pass

    elif desc_table_type == desc_table_type.AVG_WITH_POSITIVE_PROB:
        descriptive_table = descriptive_table.drop(PerfStat.AVG.to_str(), axis=1)
        descriptive_table = descriptive_table.drop(PerfStat.STD.to_str(), axis=1)

        positive = np.where(np.greater(data_np, 0.0), 1.0, 0.0)
        prob = np.sum(positive, axis=0) / data_np.shape[0]
        descriptive_table[PerfStat.POSITIVE.to_str(short=True, short_n=True)] = ['{:.1%}'.format(x) for x in prob]

    elif desc_table_type == desc_table_type.WITH_POSITIVE_PROB:
        positive = np.where(np.greater(data_np, 0.0), 1.0, 0.0)
        prob = np.sum(positive, axis=0) / data_np.shape[0]
        descriptive_table[PerfStat.POSITIVE.to_str(short=True, short_n=True)] = ['{:.1%}'.format(x) for x in prob]

    elif desc_table_type == desc_table_type.WITH_KURTOSIS:
        descriptive_table[PerfStat.SKEWNESS.to_str(short=True, short_n=True)] = [norm_variable_display_type.format(x) for x in skew(data_np, axis=0, nan_policy=nan_policy)]
        descriptive_table[PerfStat.KURTOSIS.to_str(short=True, short_n=True)] = [norm_variable_display_type.format(x) for x in kurtosis(data_np, axis=0, nan_policy=nan_policy)]
        k2, ps = normaltest(a=data_np, axis=0, nan_policy='omit')
        descriptive_table[PerfStat.NORMTEST.to_str(short=True, short_n=True)] = ['{:.2f}'.format(x) for x in ps]

    elif desc_table_type == desc_table_type.SKEW_KURTOSIS:
        descriptive_table = descriptive_table.drop(PerfStat.AVG.to_str(), axis=1)
        descriptive_table = descriptive_table.drop(PerfStat.STD.to_str(), axis=1)
        descriptive_table[PerfStat.SKEWNESS.to_str(short=True, short_n=True)] = [norm_variable_display_type.format(x) for x in skew(data_np, axis=0, nan_policy=nan_policy)]
        descriptive_table[PerfStat.KURTOSIS.to_str(short=True, short_n=True)] = [norm_variable_display_type.format(x) for x in kurtosis(data_np, axis=0, nan_policy=nan_policy)]
    elif desc_table_type == desc_table_type.WITH_SCORE:
        column_data = [df[column].dropna() for column in df.columns]
        percentiles = [percentileofscore(a=x, score=x.iloc[-1], kind='rank') for x in column_data]
        descriptive_table[PerfStat.LAST.to_str()] = [var_format.format(x.iloc[-1]) for x in column_data]
        descriptive_table[PerfStat.RANK.to_str()] = ['{:.0%}'.format(0.01*x) for x in percentiles]

    elif desc_table_type == desc_table_type.EXTENSIVE:
        descriptive_table[PerfStat.SKEWNESS.to_str(short=True, short_n=True)] \
            = [norm_variable_display_type.format(x) for x in skew(df.values, axis=0, nan_policy=nan_policy)]
        descriptive_table[PerfStat.KURTOSIS.to_str(short=True, short_n=True)] \
            = [norm_variable_display_type.format(x) for x in kurtosis(df.values, axis=0, nan_policy=nan_policy)]
        descriptive_table[PerfStat.MIN.to_str()] \
            = [var_format.format(x) for x in np.nanmin(df.values, axis=0)]
        descriptive_table[PerfStat.QUANT_M_1STD.to_str(short=True)]\
            = [var_format.format(x) for x in np.nanquantile(df.values, q=0.16, axis=0)]
        descriptive_table[PerfStat.MEDIAN.to_str(short=True)] \
            = [var_format.format(x) for x in np.nanmedian(df.values, axis=0)]
        descriptive_table[PerfStat.QUANT_P1_STD.to_str(short=True)] \
            = [var_format.format(x) for x in np.nanquantile(df.values, q=0.84, axis=0)]
        descriptive_table[PerfStat.MAX.to_str()] \
            = [var_format.format(x) for x in np.nanmax(df.values, axis=0)]

    else:
        raise TypeError(f"desc_table_type={desc_table_type} is not implemented")

    return descriptive_table


class UnitTests(Enum):
    TABLE = 1


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    returns = load_etf_data().dropna().asfreq('Q').pct_change()

    if unit_test == UnitTests.TABLE:
        df = compute_desc_table(df=returns,
                                desc_table_type=DescTableType.WITH_KURTOSIS,
                                var_format='{:.2f}',
                                annualize_vol=True,
                                is_add_tstat=False)
        print(df)


if __name__ == '__main__':

    unit_test = UnitTests.TABLE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
plot histogram
"""
# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from scipy import stats as stats
from scipy.stats import norm
from statsmodels import api as sm
from typing import Union, List, Optional, Tuple
from enum import Enum

# qis
import qis.plots.utils as put
import qis.perfstats.desc_table as dsc


class PdfType(Enum):
    KDE = 1
    KDE_NORM = 2
    HISTOGRAM = 3
    TRUNCETED_PDF = 4


def plot_histogram(df: Union[pd.DataFrame, pd.Series],
                   pdf_type: PdfType = PdfType.KDE,
                   is_drop_na: bool = True,
                   title: str = None,
                   colors: List[str] = None,
                   legend_stats: put.LegendStats = put.LegendStats.NONE,
                   add_norm_std_pdf: bool = False,
                   add_data_std_pdf: bool = False,
                   bbox_to_anchor: Optional[Tuple[float, float]] = None,
                   legend_loc: Optional[str] = 'upper left',
                   xlabel: str = None,
                   ylabel: str = None,
                   xvar_format: str = '{:.2f}',
                   yvar_format: Optional[str] = None,
                   fontsize: int = 10,
                   linewidth: float = 1.0,
                   x_limits: Tuple[Optional[float], Optional[float]] = None,
                   y_limits: Tuple[Optional[float], Optional[float]] = (0.0, None),
                   x_min_max_quantiles: Tuple[Optional[float], Optional[float]] = None,
                   clip: Tuple[Optional[float], Optional[float]] = None,
                   add_last_value: bool = False,
                   add_total_sample_pdf: bool = False,  # concat total
                   total_sample_name: str = 'Universe',
                   annualize_vol: bool = False,
                   desc_table_type: Optional[dsc.DescTableType] = dsc.DescTableType.SHORT,
                   first_color_fixed: bool = False,
                   fill: bool = False,
                   bins: Optional[Union[np.ndarray, int]] = None,
                   ax: plt.Subplot = None,
                   **kwargs
                   ) -> plt.Figure:

    df = df.copy()

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if isinstance(df, pd.DataFrame):
        n = len(df.columns)
    elif isinstance(df, pd.Series):
        df = df.to_frame()
        n = 1
    else:
        raise TypeError(f"unsupported data type = {type(df)}")

    if add_total_sample_pdf:
        first_color_fixed = True
        n = n + 1
        # unstack and remove all index data
        universe_data = df.unstack().reset_index(drop=True).rename(total_sample_name)
        # index is not encessary, nan data will be removed in loop
        df = pd.concat([universe_data,
                        df.reset_index(drop=True)
                        ], axis=1)

    if colors is None:
        colors = put.get_n_colors(n=n,
                                  first_color_fixed=first_color_fixed,
                                  **kwargs)

    for column, color in zip(df.columns, colors):
        column_data = df[column]
        if is_drop_na:  # only drop column-wise
            column_data = column_data.dropna()
        if column_data.empty:
            print(f"in plot_pdf: column{column} is empty")
            continue

        column_data = column_data.to_numpy()

        if x_min_max_quantiles is not None:
            if x_min_max_quantiles[0] is not None:
                x_min = np.nanquantile(column_data, q= x_min_max_quantiles[0])
                column_data = column_data[column_data >x_min]
            if x_min_max_quantiles[1] is not None:
                x_max = np.nanquantile(column_data, q=x_min_max_quantiles[1])
                column_data = column_data[column_data < x_max]

        if pdf_type == PdfType.KDE:
            sns.kdeplot(column_data, fill=fill, color=color, linewidth=linewidth,
                        clip=clip,
                        ax=ax)

        elif pdf_type == PdfType.KDE_NORM:
            x, y = kde_dens(column_data)
            ax.plot(x, y, color=color, linewidth=linewidth)

        elif pdf_type == PdfType.HISTOGRAM:
            sns.distplot(a=column_data, hist=True, kde=False, color=color,
                         norm_hist=True,
                         bins=bins,
                         ax=ax)

        elif pdf_type == PdfType.TRUNCETED_PDF:
            x, y = trunc_dens(column_data)
            ax.plot(x, y, color=color, linewidth=linewidth)

    if add_last_value:
        ymin, ymax = ax.get_ylim()
        hatches = put.get_n_hatch(n=len(df.columns))
        dy = (ymax-ymin) / (len(df.columns) + 1.0)
        y_positions = [0.9 * ymax - n * dy for n in range(len(df.columns))]
        xmin, xmax = ax.get_xlim()
        dx = 0.5 / (xmax-xmin)
        for idx, column in enumerate(df.columns):
            column_data = df[column].to_numpy()
            rects = ax.bar(column_data[-1], y_positions[idx], dx, color=colors[idx], hatch=hatches[idx], alpha=0.5)
            label = f"{column}\nlast={xvar_format.format(column_data[-1])}"
            percentile = stats.percentileofscore(a=column_data, score=column_data[-1], kind='rank')
            label = f"{label}\nrank={'{:.0%}'.format(0.01*percentile)}"
            put.autolabel(ax=ax, rects=rects, xpos='right', label0=label, color=colors[idx], fontsize=fontsize)

    if add_norm_std_pdf or add_data_std_pdf:
        xmin, xmax = ax.get_xlim()
        if add_data_std_pdf:
            norm_mean = np.mean(np.nanmean(df.to_numpy(), axis=0))
            norm_std = np.sqrt(np.mean(np.nanvar(df.to_numpy(), axis=0)))
        else:
            norm_mean = 0.0
            norm_std = 1.0

        dx = 0.1 * norm_std
        x_axis = np.arange(xmin, xmax, dx)
        ax.plot(x_axis, norm.pdf(x_axis, norm_mean, norm_std),
                label=f"Normal PDF ({xvar_format.format(norm_mean)},  {xvar_format.format(norm_std)})",
                color='black',
                lw=1, linestyle='--')
        ymin, ymax = ax.get_ylim()
        ax.set_ylim([0.0, ymax])

    if y_limits is not None:
        put.set_y_limits(ax=ax, y_limits=y_limits)
    if x_limits is not None:
        put.set_x_limits(ax=ax, x_limits=x_limits)

    if desc_table_type is not None and desc_table_type != dsc.DescTableType.NONE:
        stats_table = dsc.compute_desc_table(df=df,
                                         annualize_vol=annualize_vol,
                                         desc_table_type=desc_table_type,
                                         var_format=xvar_format,
                                         **kwargs)
        put.set_legend_with_stats_table(stats_table=stats_table,
                                        ax=ax,
                                        colors=colors,
                                        legend_loc=legend_loc,
                                        bbox_to_anchor=bbox_to_anchor,
                                        fontsize=fontsize,
                                        **kwargs)
    else:
        legend_labels = put.get_legend_lines(data=df,
                                             legend_stats=legend_stats,
                                             var_format=xvar_format)
        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       legend_loc=legend_loc,
                       bbox_to_anchor=bbox_to_anchor,
                       fontsize=fontsize,
                       **kwargs)
    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    # cannot use for sns.distplot
    # put.set_ax_tick_labels(ax=ax, fontsize=fontsize, skip_y_axis=True, **kwargs)
    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, fontsize=fontsize, **kwargs)
    put.set_ax_ticks_format(ax=ax, xvar_format=xvar_format, yvar_format=yvar_format, fontsize=fontsize, **kwargs)
    if yvar_format is None:
        ax.set_yticklabels('')
    put.set_spines(ax=ax, **kwargs)

    return fig


def kde_dens(x: np.ndarray, bandwidth: float = 0.5) -> Tuple[np.ndarray, np.ndarray]:
    bandwidth = bandwidth*np.nanstd(x)
    kde = sm.nonparametric.KDEUnivariate(x)
    kde.fit(bw=bandwidth)
    h = kde.bw
    d = sm.nonparametric.KDEUnivariate(x)
    d = d.fit(bw=h, fft=False)
    d_support = d.support
    d_dens = d.density
    d_dens = d_dens / np.nansum(d_dens)
    return d_support, d_dens


def trunc_dens(x: np.ndarray, bandwidth: float = 0.5) -> Tuple[np.ndarray, np.ndarray]:
    """
    with truncation at zero:
    """
    bandwidth = bandwidth*np.nanstd(x)
    kde = sm.nonparametric.KDEUnivariate(x)
    kde.fit(bw=bandwidth)
    h = kde.bw
    w = 1/(1-norm.cdf(0, loc=x, scale=h))
    d = sm.nonparametric.KDEUnivariate(x)
    d = d.fit(bw=h, weights=w / len(x), fft=False)
    d_support = d.support
    d_dens = d.density
    d_dens[d_support < 0] = 0
    d_dens = d_dens / np.nansum(d_dens)
    return d_support, d_dens


class UnitTests(Enum):
    TEST = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.TEST:
        np.random.seed(1)
        n_instruments = 100
        m_samples = 2
        exposures_nm = np.random.normal(0.0, 1.0, size=(n_instruments, m_samples))
        data = pd.DataFrame(data=exposures_nm, columns=[f"id{n+1}" for n in range(m_samples)])

        fig, ax = plt.subplots(1, 1, figsize=(3.9, 3.4), tight_layout=True)
        global_kwargs = dict(fontsize=6, linewidth=0.5, weight='normal', first_color_fixed=True)

        plot_histogram(df=data,
                       add_data_std_pdf=True,
                       add_last_value=True,
                       ax=ax,
                       **global_kwargs)
        # ax.locator_params(nbins=10, axis='x')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.TEST

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import yfinance as yf
from typing import Tuple, List, Dict
from enum import Enum

# qis
import qis
from qis.portfolio.optimization.qp_solvers import PortfolioObjective, max_portfolio_sharpe_qp
import qis.portfolio.backtester as bp
import qis.portfolio.optimization.rolling_portfolios as rlp


class UnitTests(Enum):
    MAX_DIVERSIFICATION = 1


def run_unit_test(unit_test: UnitTests):

    # data
    tickers = ['SPY', 'QQQ', 'EEM', 'TLT', 'IEF', 'LQD', 'HYG', 'SHY', 'GLD']
    prices = yf.download(tickers, start=None, end=None)['Adj Close'].dropna()

    kwargs = dict(add_mean_levels=True,
                  is_yaxis_limit_01=True,
                  baseline='zero',
                  bbox_to_anchor=(0.4, 1.1),
                  legend_stats=qis.LegendStats.AVG_STD_LAST,
                  ncol=len(prices.columns)//3,
                  var_format='{:.0%}')

    if unit_test == UnitTests.MAX_DIVERSIFICATION:
        weights = rlp.solve_optimal_weights_ewm_covar(prices=prices,
                                                      portfolio_objective=PortfolioObjective.MAX_DIVERSIFICATION,
                                                      weight_mins=np.zeros(len(prices.columns)),
                                                      weight_maxs=np.ones(len(prices.columns)),
                                                      rebalancing_freq='Q',
                                                      ewm_lambda=0.97)

        portfolio_out = bp.backtest_model_portfolio(prices=prices,
                                                    weights=weights,
                                                    is_rebalanced_at_first_date=True,
                                                    ticker='MaxDiversification',
                                                    is_output_portfolio_data=True)


        portfolio_out.plot_nav()
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MAX_DIVERSIFICATION

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
